# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import dataclasses
import enum
from collections.abc import Callable, Generator
from typing import Any, cast

from tapl_lang.core import syntax, tapl_error

################################################################################
# Python AST Terms
#
# These terms closely mirror the classes in the `ast` module.
# Keep them sorted as in https://docs.python.org/3/library/ast.html
################################################################################

Identifier = str | Callable[[syntax.BackendSetting], str]


@dataclasses.dataclass
class Module(syntax.Term):
    body: list[syntax.Term]

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from self.body

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(lambda layer: Module(body=[layer(b) for b in self.body]))


################################################################################
# STATEMENTS


@dataclasses.dataclass
class FunctionDef(syntax.Term):
    name: Identifier
    posonlyargs: list[str]
    args: list[str]
    vararg: str | None
    kwonlyargs: list[str]
    kw_defaults: list[syntax.Term]
    kwarg: str | None
    defaults: list[syntax.Term]
    body: syntax.Term
    decorator_list: list[syntax.Term]
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from self.kw_defaults
        yield from self.defaults
        yield self.body
        yield from self.decorator_list

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: FunctionDef(
                name=self.name,
                posonlyargs=self.posonlyargs,
                args=self.args,
                vararg=self.vararg,
                kwonlyargs=self.kwonlyargs,
                kw_defaults=[layer(k) for k in self.kw_defaults],
                kwarg=self.kwarg,
                defaults=[layer(d) for d in self.defaults],
                body=layer(self.body),
                decorator_list=[layer(d) for d in self.decorator_list],
                location=self.location,
            )
        )


@dataclasses.dataclass
class ClassDef(syntax.Term):
    name: Identifier
    bases: list[syntax.Term]
    keywords: list[tuple[str, syntax.Term]]
    body: syntax.Term
    decorator_list: list[syntax.Term]
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from self.bases
        yield from (t for _, t in self.keywords)
        yield self.body
        yield from self.decorator_list

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: ClassDef(
                name=self.name,
                bases=[layer(b) for b in self.bases],
                keywords=[(k, layer(v)) for k, v in self.keywords],
                body=layer(self.body),
                decorator_list=[layer(d) for d in self.decorator_list],
                location=self.location,
            )
        )


@dataclasses.dataclass
class Return(syntax.Term):
    value: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        if self.value is None:
            raise ValueError('Return statement must have a value')
        yield self.value

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(lambda layer: Return(value=layer(self.value), location=self.location))


@dataclasses.dataclass
class Delete(syntax.Term):
    targets: list[syntax.Term]
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from self.targets

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(lambda layer: Delete(targets=[layer(t) for t in self.targets], location=self.location))


@dataclasses.dataclass
class Assign(syntax.Term):
    targets: list[syntax.Term]
    value: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from self.targets
        yield self.value

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: Assign(
                targets=[layer(t) for t in self.targets], value=layer(self.value), location=self.location
            )
        )


@dataclasses.dataclass
class For(syntax.Term):
    target: syntax.Term
    iter: syntax.Term
    body: syntax.Term
    orelse: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.target
        yield self.iter
        yield self.body
        yield self.orelse

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: For(
                target=layer(self.target),
                iter=layer(self.iter),
                body=layer(self.body),
                orelse=layer(self.orelse),
                location=self.location,
            )
        )


@dataclasses.dataclass
class While(syntax.Term):
    test: syntax.Term
    body: syntax.Term
    orelse: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.test
        yield self.body
        yield self.orelse

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: While(
                test=layer(self.test),
                body=layer(self.body),
                orelse=layer(self.orelse),
                location=self.location,
            )
        )


@dataclasses.dataclass
class If(syntax.Term):
    test: syntax.Term
    body: syntax.Term
    orelse: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.test
        yield self.body
        yield self.orelse

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: If(
                test=layer(self.test),
                body=layer(self.body),
                orelse=layer(self.orelse),
                location=self.location,
            )
        )


@dataclasses.dataclass
class WithItem(syntax.Term):
    context_expr: syntax.Term
    optional_vars: syntax.Term

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.context_expr
        yield self.optional_vars

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: WithItem(
                context_expr=layer(self.context_expr),
                optional_vars=layer(self.optional_vars),
            )
        )


@dataclasses.dataclass
class With(syntax.Term):
    items: list[syntax.Term]
    body: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from self.items
        yield self.body

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: With(
                items=[layer(i) for i in self.items],
                body=layer(self.body),
                location=self.location,
            )
        )


@dataclasses.dataclass
class Raise(syntax.Term):
    exception: syntax.Term
    cause: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.exception
        yield self.cause

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: Raise(
                exception=layer(self.exception),
                cause=layer(self.cause),
                location=self.location,
            )
        )


@dataclasses.dataclass
class Try(syntax.Term):
    body: syntax.Term
    handlers: list[syntax.Term]
    orelse: syntax.Term
    finalbody: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.body
        yield from self.handlers
        yield self.orelse
        yield self.finalbody

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: Try(
                body=layer(self.body),
                handlers=[layer(h) for h in self.handlers],
                orelse=layer(self.orelse),
                finalbody=layer(self.finalbody),
                location=self.location,
            )
        )


@dataclasses.dataclass
class ExceptHandler(syntax.Term):
    exception_type: syntax.Term
    name: Identifier | None
    body: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.exception_type
        yield self.body

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: ExceptHandler(
                exception_type=layer(self.exception_type),
                name=self.name,
                body=layer(self.body),
                location=self.location,
            )
        )


@dataclasses.dataclass
class Alias:
    name: str
    asname: str | None = None


@dataclasses.dataclass
class Import(syntax.Term):
    names: list[Alias]
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from ()

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda _: Import(names=[Alias(name=n.name, asname=n.asname) for n in self.names], location=self.location)
        )


@dataclasses.dataclass
class ImportFrom(syntax.Term):
    module: str | None
    names: list[Alias]
    level: int
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from ()

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda _: ImportFrom(
                module=self.module,
                names=[Alias(name=n.name, asname=n.asname) for n in self.names],
                level=self.level,
                location=self.location,
            )
        )


@dataclasses.dataclass
class Expr(syntax.Term):
    value: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.value

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(lambda layer: Expr(value=layer(self.value), location=self.location))


@dataclasses.dataclass
class Pass(syntax.Term):
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from ()

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(lambda _: Pass(location=self.location))


################################################################################
# EXPRESSIONS


@dataclasses.dataclass
class BoolOp(syntax.Term):
    operator: str
    values: list[syntax.Term]
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from self.values

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: BoolOp(operator=self.operator, values=[layer(v) for v in self.values], location=self.location)
        )


# FIXME: target of ast.NamedExpr accepts only ast.Name. This prevents us to assign attributes like s0.name := s0.Int. Figure out how to support that.
@dataclasses.dataclass
class NamedExpr(syntax.Term):
    target: syntax.Term
    value: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.target
        yield self.value

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: NamedExpr(
                target=layer(self.target),
                value=layer(self.value),
                location=self.location,
            )
        )


@dataclasses.dataclass
class BinOp(syntax.Term):
    left: syntax.Term
    op: str
    right: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.left
        yield self.right

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: BinOp(left=layer(self.left), op=self.op, right=layer(self.right), location=self.location)
        )


@dataclasses.dataclass
class UnaryOp(syntax.Term):
    op: str
    operand: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.operand

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(lambda layer: UnaryOp(op=self.op, operand=layer(self.operand), location=self.location))


@dataclasses.dataclass
class Set(syntax.Term):
    elements: list[syntax.Term]
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from self.elements

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(lambda layer: Set(elements=[layer(v) for v in self.elements], location=self.location))


@dataclasses.dataclass
class Dict(syntax.Term):
    keys: list[syntax.Term]
    values: list[syntax.Term]
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from self.keys
        yield from self.values

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: Dict(
                keys=[layer(k) for k in self.keys],
                values=[layer(v) for v in self.values],
                location=self.location,
            )
        )


@dataclasses.dataclass
class Compare(syntax.Term):
    left: syntax.Term
    operators: list[str]
    comparators: list[syntax.Term]
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.left
        yield from self.comparators

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: Compare(
                left=layer(self.left),
                operators=self.operators,
                comparators=[layer(v) for v in self.comparators],
                location=self.location,
            )
        )


@dataclasses.dataclass
class Call(syntax.Term):
    func: syntax.Term
    args: list[syntax.Term]
    keywords: list[tuple[str, syntax.Term]]
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.func
        yield from self.args
        yield from (v for _, v in self.keywords)

    def separate(self, ls) -> list[syntax.Term]:
        return ls.build(
            lambda layer: Call(
                func=layer(self.func),
                args=[layer(v) for v in self.args],
                keywords=[(k, layer(v)) for k, v in self.keywords],
                location=self.location,
            )
        )


@dataclasses.dataclass
class Constant(syntax.Term):
    value: Any
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from ()

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(lambda _: Constant(value=self.value, location=self.location))


@dataclasses.dataclass
class Attribute(syntax.Term):
    value: syntax.Term
    attr: Identifier
    ctx: str
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.value

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: Attribute(value=layer(self.value), attr=self.attr, ctx=self.ctx, location=self.location)
        )


@dataclasses.dataclass
class Subscript(syntax.Term):
    value: syntax.Term
    slice: syntax.Term
    ctx: str
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.value
        yield self.slice

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: Subscript(
                value=layer(self.value),
                slice=layer(self.slice),
                ctx=self.ctx,
                location=self.location,
            )
        )


@dataclasses.dataclass
class Name(syntax.Term):
    id: Identifier
    ctx: str
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from ()

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(lambda _: Name(id=self.id, ctx=self.ctx, location=self.location))


@dataclasses.dataclass
class List(syntax.Term):
    elements: list[syntax.Term]
    ctx: str
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from self.elements

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: List(elements=[layer(v) for v in self.elements], ctx=self.ctx, location=self.location)
        )


@dataclasses.dataclass
class Tuple(syntax.Term):
    elements: list[syntax.Term]
    ctx: str
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from self.elements

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: Tuple(elements=[layer(v) for v in self.elements], ctx=self.ctx, location=self.location)
        )


@dataclasses.dataclass
class Slice(syntax.Term):
    lower: syntax.Term
    upper: syntax.Term
    step: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.lower
        yield self.upper
        yield self.step

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: Slice(
                lower=layer(self.lower),
                upper=layer(self.upper),
                step=layer(self.step),
                location=self.location,
            )
        )


################################################################################
# Untyped Terms
#
# These terms extend the Python AST terms defined above.
################################################################################


@dataclasses.dataclass
class Select(syntax.Term):
    value: syntax.Term
    names: list[str]
    ctx: str
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.value

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: Select(
                value=layer(self.value),
                names=self.names,
                ctx=self.ctx,
                location=self.location,
            )
        )

    def unfold(self) -> syntax.Term:
        if not self.names:
            return syntax.ErrorTerm(
                message='At least one name is required to select a path.',
                location=self.location,
            )
        value = self.value
        for i in range(len(self.names) - 1):
            value = Attribute(
                value=value,
                attr=self.names[i],
                ctx='load',
                location=self.location,
            )
        return Attribute(
            value=value,
            attr=self.names[-1],
            ctx=self.ctx,
            location=self.location,
        )


@dataclasses.dataclass
class Path(syntax.Term):
    names: list[str]
    # FIXME: Find a better name for the ctx field. options: context, reference_mode. Or keep it as it is since it's already used in the Python AST and has a clear meaning.
    ctx: str
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: Path(names=self.names, ctx=self.ctx, mode=layer(self.mode), location=self.location)
        )

    def unfold(self) -> syntax.Term:
        if len(self.names) <= 1:
            return syntax.ErrorTerm(message='At least two names are required to create a path.', location=self.location)
        value: syntax.Term = TypedName(location=self.location, id=self.names[0], ctx='load', mode=self.mode)
        for i in range(1, len(self.names) - 1):
            value = Attribute(value=value, attr=self.names[i], ctx='load', location=self.location)
        return Attribute(value=value, attr=self.names[-1], ctx=self.ctx, location=self.location)


@dataclasses.dataclass
class BranchTyping(syntax.Term):
    branches: list[syntax.Term]
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from self.branches

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(lambda layer: BranchTyping(branches=[layer(b) for b in self.branches], location=self.location))

    def unfold(self) -> syntax.Term:
        def nested_scope(inner_term: syntax.Term) -> syntax.Term:
            return syntax.BackendSettingTerm(
                backend_setting_changer=syntax.BackendSettingChanger(
                    lambda setting: setting.clone(scope_level=setting.scope_level + 1)
                ),
                term=inner_term,
            )

        def new_scope() -> syntax.Term:
            return Assign(
                targets=[
                    nested_scope(
                        Name(
                            id=lambda setting: setting.scope_name,
                            ctx='store',
                            location=self.location,
                        )
                    )
                ],
                value=Call(
                    func=Path(
                        names=['tapl_typing', 'fork_scope'],
                        ctx='load',
                        mode=MODE_TYPECHECK_NO_SCOPE,
                        location=self.location,
                    ),
                    args=[
                        Name(
                            id=lambda setting: setting.forker_name,
                            ctx='load',
                            location=self.location,
                        )
                    ],
                    keywords=[],
                    location=self.location,
                ),
                location=self.location,
            )

        body: list[syntax.Term] = []
        for b in self.branches:
            body.append(new_scope())
            body.append(nested_scope(b))

        return With(
            location=self.location,
            items=[
                WithItem(
                    context_expr=Call(
                        func=Path(
                            names=['tapl_typing', 'scope_forker'],
                            ctx='load',
                            mode=MODE_TYPECHECK_NO_SCOPE,
                            location=self.location,
                        ),
                        args=[
                            Name(
                                id=lambda setting: setting.scope_name,
                                ctx='load',
                                location=self.location,
                            )
                        ],
                        keywords=[],
                        location=self.location,
                    ),
                    optional_vars=Name(
                        id=lambda setting: setting.forker_name,
                        ctx='store',
                        location=self.location,
                    ),
                )
            ],
            body=syntax.TermList(terms=body),
        )


################################################################################
# Typed Terms
#
# These terms add a type layer to enable type checking.
################################################################################


@dataclasses.dataclass
class ModeTerm(syntax.Term):
    typecheck: bool = False
    use_scope: bool = False

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from ()

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(lambda _: self)


MODE_EVALUATE = ModeTerm(typecheck=False, use_scope=False)
MODE_EVALUATE_WITH_SCOPE = ModeTerm(typecheck=False, use_scope=True)
MODE_TYPECHECK = ModeTerm(typecheck=True, use_scope=True)
MODE_TYPECHECK_NO_SCOPE = ModeTerm(typecheck=True, use_scope=False)
MODE_SAFE = syntax.Layers(layers=[MODE_EVALUATE, MODE_TYPECHECK])
MODE_LIFT = syntax.Layers(layers=[MODE_EVALUATE, MODE_EVALUATE_WITH_SCOPE])
SAFE_LAYER_COUNT = len(MODE_SAFE.layers)


@dataclasses.dataclass
class LayerOnly(syntax.Term):
    layer_index: int
    term: syntax.Term

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.term

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        separated = ls.build(lambda layer: layer(self.term))
        return [separated[i] if i == self.layer_index else syntax.Empty for i in range(ls.layer_count)]

    def unfold(self) -> syntax.Term:
        return self.term


def get_mode(*, typecheck: bool, use_scope: bool) -> ModeTerm:
    if typecheck:
        return MODE_TYPECHECK if use_scope else MODE_TYPECHECK_NO_SCOPE
    return MODE_EVALUATE_WITH_SCOPE if use_scope else MODE_EVALUATE


@dataclasses.dataclass
class TypedName(syntax.Term):
    id: Identifier
    ctx: str
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: TypedName(id=self.id, ctx=self.ctx, mode=layer(self.mode), location=self.location)
        )

    def unfold(self) -> syntax.Term:
        if isinstance(self.mode, ModeTerm):
            if self.mode.use_scope:
                return Attribute(
                    value=Name(id=lambda setting: setting.scope_name, ctx='load', location=self.location),
                    attr=self.id,
                    ctx=self.ctx,
                    location=self.location,
                )
            return Name(id=self.id, ctx=self.ctx, location=self.location)
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class TypedAssign(syntax.Term):
    target_name: syntax.Term
    target_type: syntax.Term
    value: syntax.Term
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.target_name
        yield self.target_type
        yield self.value
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: TypedAssign(
                target_name=layer(self.target_name),
                target_type=layer(self.target_type),
                value=layer(self.value),
                mode=layer(self.mode),
                location=self.location,
            )
        )

    def unfold(self) -> syntax.Term:
        if self.mode is MODE_EVALUATE:
            return Assign(
                targets=[self.target_name],
                value=self.value,
                location=self.location,
            )
        if self.mode is MODE_TYPECHECK:
            expected = Assign(
                targets=[self.target_name],
                value=self.target_type,
                location=self.location,
            )
            assigned = Assign(targets=[self.target_name], value=self.value, location=self.location)
            return syntax.TermList(terms=[expected, assigned])
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class NoneLiteral(syntax.Term):
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(lambda layer: NoneLiteral(mode=layer(self.mode), location=self.location))

    def unfold(self) -> syntax.Term:
        if isinstance(self.mode, ModeTerm):
            if self.mode.typecheck:
                return TypedName(id='NoneType', ctx='load', mode=self.mode, location=self.location)
            return Constant(value=None, location=self.location)
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class BooleanLiteral(syntax.Term):
    value: bool
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: BooleanLiteral(
                value=self.value,
                mode=layer(self.mode),
                location=self.location,
            )
        )

    def unfold(self) -> syntax.Term:
        if isinstance(self.mode, ModeTerm):
            if self.mode.typecheck:
                return TypedName(id='Bool', ctx='load', mode=self.mode, location=self.location)
            return Constant(value=self.value, location=self.location)
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class IntegerLiteral(syntax.Term):
    value: int
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: IntegerLiteral(
                value=self.value,
                mode=layer(self.mode),
                location=self.location,
            )
        )

    def unfold(self) -> syntax.Term:
        if isinstance(self.mode, ModeTerm):
            if self.mode.typecheck:
                return TypedName(id='Int', ctx='load', mode=self.mode, location=self.location)
            return Constant(value=self.value, location=self.location)
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class FloatLiteral(syntax.Term):
    value: float
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: FloatLiteral(
                value=self.value,
                mode=layer(self.mode),
                location=self.location,
            )
        )

    def unfold(self) -> syntax.Term:
        if isinstance(self.mode, ModeTerm):
            if self.mode.typecheck:
                return TypedName(id='Float', ctx='load', mode=self.mode, location=self.location)
            return Constant(value=self.value, location=self.location)
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class StringLiteral(syntax.Term):
    value: str
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: StringLiteral(
                value=self.value,
                mode=layer(self.mode),
                location=self.location,
            )
        )

    def unfold(self) -> syntax.Term:
        if isinstance(self.mode, ModeTerm):
            if self.mode.typecheck:
                return TypedName(id='Str', ctx='load', mode=self.mode, location=self.location)
            return Constant(value=self.value, location=self.location)
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class TypedList(syntax.Term):
    elements: list[syntax.Term]
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from self.elements
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: TypedList(
                elements=[layer(e) for e in self.elements],
                mode=layer(self.mode),
                location=self.location,
            )
        )

    def unfold(self) -> syntax.Term:
        if self.mode is MODE_EVALUATE:
            return List(elements=self.elements, ctx='load', location=self.location)
        if self.mode is MODE_TYPECHECK:
            return Call(
                func=Path(
                    names=['tapl_typing', 'create_typed_list'],
                    ctx='load',
                    mode=MODE_TYPECHECK_NO_SCOPE,
                    location=self.location,
                ),
                args=self.elements,
                keywords=[],
                location=self.location,
            )
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class TypedSet(syntax.Term):
    elements: list[syntax.Term]
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from self.elements
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: TypedSet(
                elements=[layer(e) for e in self.elements],
                mode=layer(self.mode),
                location=self.location,
            )
        )

    def unfold(self) -> syntax.Term:
        if self.mode is MODE_EVALUATE:
            return Set(
                elements=self.elements,
                location=self.location,
            )
        if self.mode is MODE_TYPECHECK:
            return Call(
                func=Path(
                    names=['tapl_typing', 'create_typed_set'],
                    ctx='load',
                    mode=MODE_TYPECHECK_NO_SCOPE,
                    location=self.location,
                ),
                args=self.elements,
                keywords=[],
                location=self.location,
            )
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class TypedDict(syntax.Term):
    keys: list[syntax.Term]
    values: list[syntax.Term]
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from self.keys
        yield from self.values
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: TypedDict(
                keys=[layer(k) for k in self.keys],
                values=[layer(v) for v in self.values],
                mode=layer(self.mode),
                location=self.location,
            )
        )

    def unfold(self) -> syntax.Term:
        if self.mode is MODE_EVALUATE:
            return Dict(
                keys=self.keys,
                values=self.values,
                location=self.location,
            )
        if self.mode is MODE_TYPECHECK:
            return Call(
                func=Path(
                    names=['tapl_typing', 'create_typed_dict'],
                    ctx='load',
                    mode=MODE_TYPECHECK_NO_SCOPE,
                    location=self.location,
                ),
                args=[
                    List(
                        elements=self.keys,
                        ctx='load',
                        location=self.location,
                    ),
                    List(
                        elements=self.values,
                        ctx='load',
                        location=self.location,
                    ),
                ],
                keywords=[],
                location=self.location,
            )
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class BoolNot(syntax.Term):
    operand: syntax.Term
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.operand
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: BoolNot(operand=layer(self.operand), mode=layer(self.mode), location=self.location)
        )

    def unfold(self) -> syntax.Term:
        if self.mode is MODE_EVALUATE:
            return UnaryOp(
                op='not',
                operand=self.operand,
                location=self.location,
            )
        if self.mode is MODE_TYPECHECK:
            # unary not operator always returns Bool type
            return TypedName(
                id='Bool',
                ctx='load',
                mode=MODE_TYPECHECK,
                location=self.location,
            )
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class TypedBoolOp(syntax.Term):
    operator: str
    values: list[syntax.Term]
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from self.values
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: TypedBoolOp(
                operator=self.operator,
                values=[layer(v) for v in self.values],
                mode=layer(self.mode),
                location=self.location,
            )
        )

    def unfold(self) -> syntax.Term:
        if self.mode is MODE_EVALUATE:
            return BoolOp(operator=self.operator, values=self.values, location=self.location)
        if self.mode is MODE_TYPECHECK:
            return Call(
                func=Path(
                    names=['tapl_typing', 'create_union'],
                    ctx='load',
                    mode=MODE_TYPECHECK_NO_SCOPE,
                    location=self.location,
                ),
                args=self.values,
                keywords=[],
                location=self.location,
            )
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class TypedReturn(syntax.Term):
    value: syntax.Term
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.value
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: TypedReturn(value=layer(self.value), mode=layer(self.mode), location=self.location)
        )

    def unfold(self) -> syntax.Term:
        if self.mode is MODE_EVALUATE:
            return Return(
                value=self.value,
                location=self.location,
            )
        if self.mode is MODE_TYPECHECK:
            call = Call(
                func=Path(
                    names=['tapl_typing', 'add_return_type'],
                    ctx='load',
                    mode=MODE_TYPECHECK_NO_SCOPE,
                    location=self.location,
                ),
                args=[
                    Name(
                        id=lambda setting: setting.scope_name,
                        ctx='load',
                        location=self.location,
                    ),
                    self.value,
                ],
                keywords=[],
                location=self.location,
            )
            return Expr(
                value=call,
                location=self.location,
            )
        raise tapl_error.UnhandledError


class ParamCategory(enum.Enum):
    POSITIONAL_ONLY = 'positional_only'
    REGULAR = 'regular'
    VAR_POSITIONAL = 'var_positional'
    KEYWORD_ONLY = 'keyword_only'
    VAR_KEYWORD = 'var_keyword'


@dataclasses.dataclass
class Parameter(syntax.Term):
    name: str
    type_: syntax.Term
    default: syntax.Term
    mode: syntax.Term
    category: ParamCategory
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.type_
        yield self.default
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: Parameter(
                name=self.name,
                type_=layer(self.type_),
                default=layer(self.default),
                mode=layer(self.mode),
                category=self.category,
                location=self.location,
            )
        )

    def unfold(self) -> syntax.Term:
        if self.mode is MODE_TYPECHECK:
            return self.type_
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class TypedFunctionDef(syntax.Term):
    name: str
    parameters: list[syntax.Term]
    return_type: syntax.Term
    body: syntax.Term
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from self.parameters
        yield self.return_type
        yield self.body
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: TypedFunctionDef(
                name=self.name,
                return_type=layer(self.return_type),
                parameters=[layer(p) for p in self.parameters],
                body=layer(self.body),
                mode=layer(self.mode),
                location=self.location,
            )
        )

    def unfold_evaluate(self) -> syntax.Term:
        if not all(cast('Parameter', p).type_ is syntax.Empty for p in self.parameters):
            raise tapl_error.TaplError(
                'All parameter types must be a syntax.Empty when generating a function in evaluate mode.'
            )

        return FunctionDef(
            name=self.name,
            posonlyargs=[],
            args=[cast('Parameter', p).name for p in self.parameters],
            vararg=None,
            kwonlyargs=[],
            kw_defaults=[],
            kwarg=None,
            defaults=[],
            body=self.body,
            decorator_list=[],
            location=self.location,
        )

    def unfold_typecheck_main(self, *, is_method: bool = False) -> syntax.Term:
        def nested_scope(nested_term: syntax.Term) -> syntax.Term:
            return syntax.BackendSettingTerm(
                backend_setting_changer=syntax.BackendSettingChanger(
                    lambda setting: setting.clone(scope_level=setting.scope_level + 1)
                ),
                term=nested_term,
            )

        param_names = [cast('Parameter', p).name for p in self.parameters]

        keywords: list[tuple[str, syntax.Term]] = []
        keywords.append(
            (
                'parent__sa',
                Name(
                    id=lambda setting: setting.scope_name,
                    ctx='load',
                    location=self.location,
                ),
            )
        )
        keywords.extend(
            (
                name,
                Name(
                    id=name,
                    ctx='load',
                    location=self.location,
                ),
            )
            for name in param_names
        )
        new_scope = Assign(
            targets=[
                nested_scope(
                    Name(location=self.location, id=lambda setting: setting.scope_name, ctx='load'),
                )
            ],
            value=Call(
                func=Path(
                    names=['tapl_typing', 'create_scope'],
                    ctx='load',
                    mode=MODE_TYPECHECK_NO_SCOPE,
                    location=self.location,
                ),
                args=[],
                keywords=keywords,
                location=self.location,
            ),
            location=self.location,
        )
        tmp_function: syntax.Term = syntax.Empty
        set_return_type: syntax.Term = syntax.Empty
        if self.return_type is not syntax.Empty:
            params = [cast('Parameter', p).type_ for p in self.parameters]
            if is_method:
                params = params[1:]  # skip 'self' parameter type
            tmp_function = Assign(
                location=self.location,
                targets=[
                    TypedName(
                        id=self.name,
                        ctx='store',
                        mode=self.mode,
                        location=self.location,
                    )
                ],
                value=Call(
                    func=Path(
                        names=['tapl_typing', 'create_function'],
                        ctx='load',
                        mode=MODE_TYPECHECK_NO_SCOPE,
                        location=self.location,
                    ),
                    args=[
                        List(
                            elements=params,
                            ctx='load',
                            location=self.location,
                        ),
                        self.return_type,
                    ],
                    keywords=[],
                    location=self.location,
                ),
            )
            set_return_type = Expr(
                value=Call(
                    func=Path(
                        names=['tapl_typing', 'set_return_type'],
                        ctx='load',
                        mode=MODE_TYPECHECK_NO_SCOPE,
                        location=self.location,
                    ),
                    args=[
                        Name(
                            id=lambda setting: setting.scope_name,
                            ctx='load',
                            location=self.location,
                        ),
                        self.return_type,
                    ],
                    keywords=[],
                    location=self.location,
                ),
                location=self.location,
            )

        get_return_type = Return(
            value=Call(
                func=Path(
                    names=['tapl_typing', 'get_return_type'],
                    ctx='load',
                    mode=MODE_TYPECHECK_NO_SCOPE,
                    location=self.location,
                ),
                args=[
                    Name(
                        id=lambda setting: setting.scope_name,
                        ctx='load',
                        location=self.location,
                    )
                ],
                keywords=[],
                location=self.location,
            ),
            location=self.location,
        )

        return FunctionDef(
            name=self.name,
            posonlyargs=[],
            args=param_names,
            vararg=None,
            kwonlyargs=[],
            kw_defaults=[],
            kwarg=None,
            defaults=[],
            body=syntax.TermList(
                terms=[
                    new_scope,
                    nested_scope(syntax.TermList(terms=[tmp_function, set_return_type, self.body, get_return_type])),
                ]
            ),
            decorator_list=[],
            location=self.location,
        )

    def unfold_typecheck_type(self) -> syntax.Term:
        if self.parameters and all(cast('Parameter', p).type_ is syntax.Empty for p in self.parameters):
            return Assign(
                targets=[
                    TypedName(
                        id=self.name,
                        ctx='store',
                        mode=self.mode,
                        location=self.location,
                    )
                ],
                value=Name(
                    id=self.name,
                    ctx='load',
                    location=self.location,
                ),
                location=self.location,
            )
        if any(cast('Parameter', p).type_ is syntax.Empty for p in self.parameters):
            raise tapl_error.TaplError(
                'Parameter types must be either all specified or all omitted when generating a function type in typecheck mode.'
            )
        return Assign(
            targets=[
                TypedName(
                    id=self.name,
                    ctx='store',
                    mode=self.mode,
                    location=self.location,
                )
            ],
            value=Call(
                func=Path(
                    names=['tapl_typing', 'create_function'],
                    ctx='load',
                    mode=MODE_TYPECHECK_NO_SCOPE,
                    location=self.location,
                ),
                args=[
                    List(
                        elements=[cast('Parameter', p).type_ for p in self.parameters],
                        ctx='load',
                        location=self.location,
                    ),
                    Call(
                        func=Name(
                            id=self.name,
                            ctx='load',
                            location=self.location,
                        ),
                        args=[cast('Parameter', p).type_ for p in self.parameters],
                        keywords=[],
                        location=self.location,
                    ),
                ],
                keywords=[],
                location=self.location,
            ),
            location=self.location,
        )

    def unfold(self) -> syntax.Term:
        if self.mode is MODE_EVALUATE:
            return self.unfold_evaluate()
        if self.mode is MODE_TYPECHECK:
            return syntax.TermList(terms=[self.unfold_typecheck_main(is_method=False), self.unfold_typecheck_type()])
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class TypedIf(syntax.Term):
    test: syntax.Term
    body: syntax.Term
    elifs: list[tuple[syntax.Term, syntax.Term]]  # (test, body)
    orelse: syntax.Term
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.test
        yield self.body
        for test, body in self.elifs:
            yield test
            yield body
        yield self.orelse
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: TypedIf(
                test=layer(self.test),
                body=layer(self.body),
                elifs=[(layer(test), layer(body)) for test, body in self.elifs],
                orelse=layer(self.orelse),
                mode=layer(self.mode),
                location=self.location,
            )
        )

    def codegen_evaluate(self) -> syntax.Term:
        return If(
            test=self.test,
            body=self.body,
            orelse=self.orelse,
            location=self.location,
        )

    def codegen_typecheck(self) -> syntax.Term:
        true_side = syntax.TermList(
            terms=[
                Expr(
                    value=self.test,
                    location=self.location,
                ),
                self.body,
            ]
        )
        return BranchTyping(branches=[true_side, self.orelse], location=self.location)

    def unfold(self) -> syntax.Term:
        # FIXME: handle elifs
        if self.elifs:
            raise tapl_error.UnhandledError('Elif clauses are not yet supported in TypedIf unfold.')
        if self.mode is MODE_EVALUATE:
            return self.codegen_evaluate()
        if self.mode is MODE_TYPECHECK:
            return self.codegen_typecheck()
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class ElifSibling(syntax.SiblingTerm):
    test: syntax.Term
    body: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.test
        yield self.body

    def integrate_into(self, previous_siblings: list[syntax.Term]) -> None:
        term = previous_siblings[-1]
        if isinstance(term, syntax.ErrorTerm):
            return
        if not isinstance(term, TypedIf):
            error = syntax.ErrorTerm('Elif can only be integrated into If.' + repr(term), location=self.location)
            previous_siblings.append(error)
        else:
            term.elifs.append((self.test, self.body))


@dataclasses.dataclass
class ElseSibling(syntax.SiblingTerm):
    body: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.body

    def integrate_into(self, previous_siblings: list[syntax.Term]) -> None:
        term = previous_siblings[-1]
        if isinstance(term, syntax.ErrorTerm):
            return
        if not isinstance(term, TypedIf):
            error = syntax.ErrorTerm('Else can only be integrated into If.' + repr(term), location=self.location)
            previous_siblings.append(error)
        elif term.orelse is not syntax.Empty:
            error = syntax.ErrorTerm('An If statement can only have one Else clause.', location=self.location)
            previous_siblings.append(error)
        else:
            term.orelse = self.body


@dataclasses.dataclass
class TypedWhile(syntax.Term):
    test: syntax.Term
    body: syntax.Term
    orelse: syntax.Term
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.test
        yield self.body
        yield self.orelse
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: TypedWhile(
                test=layer(self.test),
                body=layer(self.body),
                orelse=layer(self.orelse),
                mode=layer(self.mode),
                location=self.location,
            )
        )

    def codegen_evaluate(self) -> syntax.Term:
        return While(
            test=self.test,
            body=self.body,
            orelse=self.orelse,
            location=self.location,
        )

    def codegen_typecheck(self) -> syntax.Term:
        return TypedIf(
            test=self.test,
            body=self.body,
            elifs=[],
            orelse=self.orelse,
            mode=self.mode,
            location=self.location,
        )

    def unfold(self) -> syntax.Term:
        if self.mode is MODE_EVALUATE:
            return self.codegen_evaluate()
        if self.mode is MODE_TYPECHECK:
            return self.codegen_typecheck()
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class TypedFor(syntax.Term):
    target: syntax.Term
    iter: syntax.Term
    body: syntax.Term
    orelse: syntax.Term
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.target
        yield self.iter
        yield self.body
        yield self.orelse
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: TypedFor(
                target=layer(self.target),
                iter=layer(self.iter),
                body=layer(self.body),
                orelse=layer(self.orelse),
                mode=layer(self.mode),
                location=self.location,
            )
        )

    def codegen_evaluate(self) -> syntax.Term:
        return For(
            target=self.target,
            iter=self.iter,
            body=self.body,
            orelse=self.orelse,
            location=self.location,
        )

    def codegen_typecheck(self) -> syntax.Term:
        iterator_type = Call(
            func=Attribute(
                value=self.iter,
                attr='__iter__',
                ctx='load',
                location=self.location,
            ),
            args=[],
            keywords=[],
            location=self.location,
        )
        item_type = Call(
            location=self.location,
            func=Attribute(
                value=iterator_type,
                attr='__next__',
                ctx='load',
                location=self.location,
            ),
            args=[],
            keywords=[],
        )
        assign_target = Assign(
            targets=[self.target],
            value=item_type,
            location=self.location,
        )
        for_branch = syntax.TermList(terms=[assign_target, self.body])
        return BranchTyping(
            branches=[for_branch, self.orelse],
            location=self.location,
        )

    def unfold(self) -> syntax.Term:
        if self.mode is MODE_EVALUATE:
            return self.codegen_evaluate()
        if self.mode is MODE_TYPECHECK:
            return self.codegen_typecheck()
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class TypedTry(syntax.Term):
    body: syntax.Term
    handlers: list[syntax.Term]
    finalbody: syntax.Term
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.body
        yield from self.handlers
        yield self.finalbody
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: TypedTry(
                body=layer(self.body),
                handlers=[layer(h) for h in self.handlers],
                finalbody=layer(self.finalbody),
                mode=layer(self.mode),
                location=self.location,
            )
        )

    def unfold(self) -> syntax.Term:
        if self.mode is MODE_EVALUATE:
            return Try(
                body=self.body,
                handlers=self.handlers,
                orelse=syntax.Empty,
                finalbody=self.finalbody,
                location=self.location,
            )
        if self.mode is MODE_TYPECHECK:
            # FIXME: Implement scope forker to preserve parent scope in try, except, and finally blocks except return type
            handlers: list[syntax.Term] = [self.body]
            for handler in self.handlers:
                if not isinstance(handler, ExceptHandler):
                    raise tapl_error.TaplError('TypedTry handlers must be ExceptHandler in type-check mode.')
                exception_name: syntax.Term = syntax.Empty
                if handler.name is not None:
                    exception_name = Assign(
                        targets=[Name(id=handler.name, ctx='store', location=self.location)],
                        value=handler.exception_type,
                        location=self.location,
                    )
                handlers.append(syntax.TermList(terms=[exception_name, handler.body]))
            handlers.append(self.finalbody)
            return BranchTyping(branches=handlers, location=self.location)
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class ExceptSibling(syntax.SiblingTerm):
    exception_type: syntax.Term
    name: str | None
    body: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.exception_type
        yield self.body

    def integrate_into(self, previous_siblings: list[syntax.Term]) -> None:
        term = previous_siblings[-1]
        if isinstance(term, syntax.ErrorTerm):
            return
        if not isinstance(term, TypedTry):
            error = syntax.ErrorTerm(
                'Except can only be integrated into typed Try.' + repr(term), location=self.location
            )
            previous_siblings.append(error)
        else:
            handler = ExceptHandler(
                exception_type=self.exception_type,
                name=self.name,
                body=self.body,
                location=self.location,
            )
            term.handlers.append(handler)


@dataclasses.dataclass
class FinallySibling(syntax.SiblingTerm):
    body: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.body

    def integrate_into(self, previous_siblings: list[syntax.Term]) -> None:
        term = previous_siblings[-1]
        if isinstance(term, syntax.ErrorTerm):
            return
        if not isinstance(term, TypedTry):
            error = syntax.ErrorTerm(
                'Finally can only be integrated into typed Try.' + repr(term), location=self.location
            )
            previous_siblings.append(error)
        elif term.finalbody is not syntax.Empty:
            error = syntax.ErrorTerm('A Try statement can only have one Finally clause.', location=self.location)
            previous_siblings.append(error)
        else:
            term.finalbody = self.body


@dataclasses.dataclass
class TypedImport(syntax.Term):
    names: list[Alias]
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: TypedImport(
                names=self.names,
                mode=layer(self.mode),
                location=self.location,
            )
        )

    def unfold(self) -> syntax.Term:
        if self.mode is MODE_EVALUATE:
            return Import(location=self.location, names=self.names)
        if self.mode is MODE_TYPECHECK:
            if len(self.names) > 1:
                raise tapl_error.TaplError(
                    'Import does not support multiple names yet.'
                )  # FIXME: Support multiple import names
            return Expr(
                value=Call(
                    location=self.location,
                    func=Path(
                        names=['tapl_typing', 'import_module'],
                        ctx='load',
                        mode=MODE_TYPECHECK_NO_SCOPE,
                        location=self.location,
                    ),
                    args=[
                        Name(
                            id=lambda setting: setting.scope_name,
                            ctx='load',
                            location=self.location,
                        ),
                        List(
                            elements=[Constant(location=self.location, value=self.names[0].name)],
                            ctx='load',
                            location=self.location,
                        ),
                    ],
                    keywords=[],
                ),
                location=self.location,
            )
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class TypedImportFrom(syntax.Term):
    module: str | None
    names: list[Alias]
    level: int
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield self.mode

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: TypedImportFrom(
                module=self.module,
                names=self.names,
                level=self.level,
                mode=layer(self.mode),
                location=self.location,
            )
        )

    def unfold(self) -> syntax.Term:
        if self.mode is MODE_EVALUATE:
            return ImportFrom(module=self.module, names=self.names, level=self.level, location=self.location)
        if self.mode is MODE_TYPECHECK:
            return Expr(
                value=Call(
                    location=self.location,
                    func=Path(
                        names=['tapl_typing', 'import_module'],
                        ctx='load',
                        mode=MODE_TYPECHECK_NO_SCOPE,
                        location=self.location,
                    ),
                    args=[
                        Name(
                            id=lambda setting: setting.scope_name,
                            ctx='load',
                            location=self.location,
                        ),
                        List(
                            elements=[Constant(location=self.location, value=name.name) for name in self.names],
                            ctx='load',
                            location=self.location,
                        ),
                        Constant(location=self.location, value=self.module),
                        Constant(location=self.location, value=self.level),
                    ],
                    keywords=[],
                ),
                location=self.location,
            )
        raise tapl_error.UnhandledError


@dataclasses.dataclass
class TypedClassDef(syntax.Term):
    name: str
    bases: list[syntax.Term]
    body: syntax.Term
    mode: syntax.Term
    location: syntax.Location

    def children(self) -> Generator[syntax.Term, None, None]:
        yield from self.bases
        yield self.body

    def separate(self, ls: syntax.LayerSeparator) -> list[syntax.Term]:
        return ls.build(
            lambda layer: TypedClassDef(
                name=self.name,
                bases=[layer(b) for b in self.bases],
                body=layer(self.body),
                mode=layer(self.mode),
                location=self.location,
            )
        )

    def codegen_evaluate(self) -> syntax.Term:
        return ClassDef(
            name=self.name,
            bases=self.bases,
            keywords=[],
            body=self.body,
            decorator_list=[],
            location=self.location,
        )

    def codegen_typecheck(self) -> syntax.Term:
        body = []
        methods: list[TypedFunctionDef] = []
        constructor_args = []
        for item in self.body.children():
            if isinstance(item, TypedFunctionDef):
                if item.name == '__init__':
                    constructor_args = item.parameters[1:]
                else:
                    methods.append(item)
                body.append(item.unfold_typecheck_main(is_method=True))
            elif isinstance(item, Assign):
                targets: list[syntax.Term] = []
                for target in item.targets:
                    if not isinstance(target, TypedName):
                        raise tapl_error.TaplError(
                            'Class variable assignments must use TypedName as target in type-check mode.'
                        )
                    targets.append(
                        TypedName(
                            id=target.id,
                            ctx='store',
                            mode=MODE_TYPECHECK_NO_SCOPE,
                            location=target.location,
                        )
                    )
                body.append(Assign(targets=targets, value=item.value, location=item.location))
            else:
                body.append(item)
        class_stmt = ClassDef(
            name=self.name,
            bases=self.bases,
            keywords=[],
            body=syntax.TermList(terms=body),
            decorator_list=[],
            location=self.location,
        )

        method_types: list[syntax.Term] = []
        for method in methods:
            # The first parameter is the instance itself, so we can set it to the instance type.
            if not (
                len(method.parameters) >= 1
                and isinstance(method.parameters[0], Parameter)
                and method.parameters[0].type_ is syntax.Empty
            ):
                raise tapl_error.TaplError(
                    f'First parameter of method {method.name} in class {self.name} must be self with no type annotation.'
                )
            tail_args = method.parameters[1:]
            method_types.append(
                Tuple(
                    elements=[
                        Constant(
                            value=method.name,
                            location=self.location,
                        ),
                        List(
                            elements=tail_args,
                            ctx='load',
                            location=self.location,
                        ),
                    ],
                    ctx='load',
                    location=self.location,
                )
            )

        create_class = Assign(
            targets=[
                TypedName(
                    id=self.name,
                    ctx='store',
                    mode=self.mode,
                    location=self.location,
                ),
            ],
            value=Call(
                func=Path(
                    names=['tapl_typing', 'create_class'],
                    ctx='load',
                    mode=MODE_TYPECHECK_NO_SCOPE,
                    location=self.location,
                ),
                args=[],
                keywords=[
                    (
                        'cls',
                        Name(
                            id=self.name,
                            ctx='load',
                            location=self.location,
                        ),
                    ),
                    (
                        'init_args',
                        List(
                            elements=constructor_args,
                            ctx='load',
                            location=self.location,
                        ),
                    ),
                    (
                        'methods',
                        List(
                            elements=method_types,
                            ctx='load',
                            location=self.location,
                        ),
                    ),
                ],
                location=self.location,
            ),
            location=self.location,
        )

        return syntax.TermList(terms=[class_stmt, create_class])

    def unfold(self) -> syntax.Term:
        if self.mode is MODE_EVALUATE:
            return self.codegen_evaluate()
        if self.mode is MODE_TYPECHECK:
            return self.codegen_typecheck()
        raise tapl_error.UnhandledError



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import argparse
import ast
import os
import pathlib

from tapl_lang.__about__ import __version__
from tapl_lang.lib.compiler import compile_tapl

# If you want to install it in editable mode for development,
# you can use the following command:
# pip install -e .


def compile_and_run(path: str) -> None:
    """
    Compiles the TAPL file at the given path.
    """
    absolute_path = os.path.abspath(path)
    p = pathlib.Path(absolute_path)
    source = p.read_text()
    layers = compile_tapl(source)
    dir_path = os.path.dirname(absolute_path)
    for i in reversed(range(len(layers))):
        suffix = i if i else ''
        filename = os.path.join(dir_path, f'{p.stem}{suffix}.py')
        python_code = ast.unparse(layers[i])
        with open(filename, 'w') as f:
            f.write(python_code)
        exec(python_code, globals())  # noqa: S102 # Find safe way to execute


def main():
    """
    Main function for the CLI application.
    """
    parser = argparse.ArgumentParser(
        prog='tapl',
        description='TAPL compiler CLI — compiles and runs .tapl source files.',
    )
    parser.add_argument(
        '-v',
        '--version',
        action='version',
        version=f'%(prog)s {__version__}',
    )
    parser.add_argument('file', type=str, help='path to a .tapl source file')
    args = parser.parse_args()

    compile_and_run(args.file)


if __name__ == '__main__':
    main()



from tapl_lang.lib import tapl_typing
from tapl_lang.pythonlike.predef1 import predef_scope as predef_scope__sa
s0 = tapl_typing.create_scope(parent__sa=predef_scope__sa)
tapl_typing.import_module(s0, ['tapl_dev'], 'tapl_lang.lib', 0)
s0.x = s0.Int
s0.x = s0.Int
s0.tapl_dev.log(s0.x)
s0.a = s0.b = s0.Str
s0.tapl_dev.log(s0.a)
s0.tapl_dev.log(s0.b)
s0.f = s0.Float
with tapl_typing.scope_forker(s0) as f0:
    s1 = tapl_typing.fork_scope(f0)
    s1.f = s1.Int / s1.Int
    s1 = tapl_typing.fork_scope(f0)
    s1.tapl_dev.log(s1.Str)
    s1 = tapl_typing.fork_scope(f0)
    s1.tapl_dev.log(s1.Str)
s0.c = tapl_typing.create_typed_list(s0.Int, s0.Int, s0.Int)
s0.tapl_dev.log(s0.c)
s0.x = s0.c[s0.Int]
s0.tapl_dev.log(s0.c[s0.Int])
s0.c[s0.Int] = s0.Int
del s0.c[s0.Int]
s0.tapl_dev.log(s0.c)
s0.s = tapl_typing.create_typed_set(s0.Int, s0.Int, s0.Int, s0.Int)
s0.tapl_dev.log(s0.s)
with tapl_typing.scope_forker(s0) as f0:
    s1 = tapl_typing.fork_scope(f0)
    s1.Int in s1.s
    s1.tapl_dev.log(s1.Str)
    s1 = tapl_typing.fork_scope(f0)
s0.s.add(s0.Int)
s0.s.remove(s0.Int)
s0.tapl_dev.log(s0.s)
s0.dict_obj = tapl_typing.create_typed_dict([s0.Str, s0.Str], [s0.Str, s0.Str])
s0.tapl_dev.log(s0.dict_obj)
s0.value = s0.dict_obj[s0.Str]
s0.tapl_dev.log(s0.dict_obj[s0.Str])
s0.dict_obj[s0.Str] = s0.Str
del s0.dict_obj[s0.Str]
s0.tapl_dev.log(s0.dict_obj)
s0.print(s0.Str)



from tapl_lang.lib import tapl_dev

def int_print(a):
    tapl_dev.log(a)
int_print(123)

def zero():
    return 0

def increment(a):
    return a + 1
int_print(increment(zero()))
if True:
    int_print(456)
a = 5
if a == 2:
    b = 7
else:
    b = 'banana'
tapl_dev.log(b)



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import importlib
import itertools
from typing import Any

from tapl_lang.lib import builtin_types as bt
from tapl_lang.lib import kinds, scope

create_union = kinds.create_union
create_function = kinds.create_function


def import_module(
    scope_: scope.Scope, names: list[str], module: str | None = None, level: int = 0, package: str | None = None
) -> None:
    for name in names:
        path = name.split('.')
        current_scope = scope_
        for p in path[:-1]:
            if not hasattr(current_scope, p):
                setattr(current_scope, p, scope.Scope(label__sa=f'{p} module'))
            current_scope = getattr(current_scope, p)
        # adding '1' suffix to module name to import the type layer
        module_name = f'{module}.{name}' if module else name
        module_name = '.' * level + module_name
        imported_module = importlib.import_module(f'{module_name}1', package=package)
        setattr(current_scope, path[-1], imported_module.s0)


def create_scope(
    parent__sa: scope.Scope | None = None,
    **kwargs: Any,
) -> scope.Scope:
    parent_scope = None
    if parent__sa:
        parent_scope = parent__sa
    current = scope.Scope(parent=parent_scope)
    current.store_many__sa(kwargs)
    return current


def set_return_type(s: scope.Scope, return_type: Any) -> None:
    if s.returns__sa or s.return_type__sa is not None:
        raise ValueError('Return type has already been set.')
    s.return_type__sa = return_type


def add_return_type(s: scope.Scope, return_type: Any) -> None:
    if s.return_type__sa is None:
        s.returns__sa.append(return_type)
    elif not kinds.check_subtype(return_type, s.return_type__sa):
        raise TypeError(f'Return type mismatch: expected {s.return_type__sa}, got {return_type}.')


def get_return_type(s: scope.Scope) -> Any:
    if s.return_type__sa is not None:
        return s.return_type__sa
    if s.returns__sa:
        return kinds.create_union(*s.returns__sa)
    return bt.NoneType


def scope_forker(s: scope.Scope) -> scope.ScopeForker:
    return scope.ScopeForker(s)


def fork_scope(forker: scope.ScopeForker) -> scope.Scope:
    return forker.new_scope()


TO_STRING_FUNCTION = kinds.Function(posonlyargs=[], args=[], result=bt.Str)


def create_class(cls, init_args: list[Any], methods: list[tuple[str, list[Any]]]) -> kinds.Function:
    fields: dict[str, Any] = {}
    class_name = getattr(cls, 'class_name', cls.__name__)
    obj_type = kinds.Record(fields=fields, label=f'{class_name}!')
    self_methods_scope = scope.Scope()
    for method_name, param_types in methods:

        def create_lazy_result(name, types):
            return lambda: getattr(cls, name)(*(obj_type, *types))

        method = kinds.create_function(args=param_types, lazy_result=create_lazy_result(method_name, param_types))
        self_methods_scope.store__sa(method_name, method)
    self_scope = scope.Scope(parent=self_methods_scope)
    cls.__init__(*[self_scope, *init_args])
    fields.update(
        {
            '__repr__': TO_STRING_FUNCTION,
            '__str__': TO_STRING_FUNCTION,
        }
    )
    for label in itertools.chain(self_methods_scope.fields__sa.keys(), self_scope.fields__sa.keys()):
        member = self_scope.load__sa(label)
        fields[label] = member

    for member in fields.values():
        if isinstance(member, kinds.Function):
            member.evaluate_lazy_result__sa()

    return kinds.Function(posonlyargs=init_args, args=[], result=obj_type)


def create_typed_list(*element_types):
    if len(element_types) == 0:
        # FIXME: Implement dynamic Any element type that can be specified at runtime (e.g., when appending Int to an empty list, element type becomes Int).
        # Maybe consider introducing an Unknown type for such cases.
        element_type = bt.Any
    elif len(element_types) == 1:
        element_type = element_types[0]
    else:
        element_type = kinds.create_union(*element_types)
    return bt.List(element_type)


def create_typed_set(*element_types):
    if len(element_types) == 0:
        element_type = bt.Any
    elif len(element_types) == 1:
        element_type = element_types[0]
    else:
        element_type = kinds.create_union(*element_types)
    return bt.Set(element_type)


def create_typed_dict(keys, values):
    if len(keys) != len(values):
        raise ValueError('Keys and values must have the same length.')
    if len(keys) == 0:
        key_type = bt.Str
        value_type = bt.Any
    elif len(keys) == 1:
        key_type, value_type = keys[0], values[0]
    else:
        key_type = kinds.create_union(*keys)
        value_type = kinds.create_union(*values)
    return bt.Dict(key_type, value_type)



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception


from tapl_lang.pythonlike import rule_names


def test_rule_name_consistency():
    assert rule_names.RULE_NAMES, 'Grammar rule names are not defined or empty.'
    for full_rule_name in rule_names.RULE_NAMES:
        names = full_rule_name.split('.')
        assert len(names) == 2, f'Rule name {full_rule_name} does not have exactly two parts.'
        assert names[0] == 'pythonlike', f'Rule name {full_rule_name} does not start with "pythonlike".'
        constant_name = names[1].upper()
        assert constant_name in rule_names.__dict__, (
            f'Rule name {constant_name} is not found as a constant in the grammar module.'
        )



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception


RULE_NAMES = set()


def declare(rule_key: str) -> str:
    if not rule_key:
        raise ValueError('Rule name cannot be empty.')
    if rule_key.lower() != rule_key:
        raise ValueError('Rule name must be in lowercase.')
    rule_name = f'pythonlike.{rule_key}'
    if rule_name in RULE_NAMES:
        raise ValueError(f'Rule {rule_name} is already defined.')
    RULE_NAMES.add(rule_name)
    return rule_name


# STARTING RULES
# ==============
START = declare('start')

# HELPER RULES
# ============
TOKEN = declare('token')

# GENERAL STATEMENTS
# ==================
STATEMENTS = declare('statements')
STATEMENT = declare('statement')
SINGLE_COMPOUND_STMT = declare('single_compound_stmt')
STATEMENT_NEWLINE = declare('statement_newline')
SIMPLE_STMTS = declare('simple_stmts')
SIMPLE_STMT = declare('simple_stmt')
COMPOUND_STMT = declare('compound_stmt')

# SIMPLE STATEMENTS
# =================
ASSIGNMENT = declare('assignment')
ANNOTATED_RHS = declare('annotated_rhs')
AUGASSIGN = declare('augassign')
RETURN_STMT = declare('return_stmt')
PASS_STMT = declare('pass_stmt')
BREAK_STMT = declare('break_stmt')
CONTINUE_STMT = declare('continue_stmt')
GLOBAL_STMT = declare('global_stmt')
NONLOCAL_STMT = declare('nonlocal_stmt')
DEL_STMT = declare('del_stmt')
YIELD_STMT = declare('yield_stmt')
ASSERT_STMT = declare('assert_stmt')
IMPORT_STMT = declare('import_stmt')
RAISE_STMT = declare('raise_stmt')

# Import statements
# -----------------
IMPORT_NAME = declare('import_name')
IMPORT_FROM = declare('import_from')
IMPORT_FROM_TARGETS = declare('import_from_targets')
IMPORT_FROM_AS_NAMES = declare('import_from_as_names')
IMPORT_FROM_AS_NAME = declare('import_from_as_name')
DOTTED_AS_NAMES = declare('dotted_as_names')
DOTTED_AS_NAME = declare('dotted_as_name')
DOTTED_NAME = declare('dotted_name')

# COMPOUND STATEMENTS
# ===================

# Common elements
# ---------------
BLOCK = declare('block')
DECORATORS = declare('decorators')

# Class definitions
# -----------------
CLASS_DEF = declare('class_def')
CLASS_DEF_RAW = declare('class_def_raw')

# Function definitions
# --------------------
FUNCTION_DEF = declare('function_def')
FUNCTION_DEF_RAW = declare('function_def_raw')

# Function parameters
# -------------------
PARAMS = declare('params')
PARAMETERS = declare('parameters')
SLASH_NO_DEFAULT = declare('slash_no_default')
SLASH_WITH_DEFAULT = declare('slash_with_default')
STAR_ETC = declare('star_etc')
KWDS = declare('kwds')
PARAM_NO_DEFAULT = declare('param_no_default')
PARAM_NO_DEFAULT_STAR_ANNOTATION = declare('param_no_default_star_annotation')
PARAM_WITH_DEFAULT = declare('param_with_default')
PARAM_MAYBE_DEFAULT = declare('param_maybe_default')
PARAM = declare('param')
PARAM_STAR_ANNOTATION = declare('param_star_annotation')
ANNOTATION = declare('annotation')
STAR_ANNOTATION = declare('star_annotation')
DEFAULT = declare('default')

# If statement
# ------------
IF_STMT = declare('if_stmt')
ELIF_STMT = declare('elif_stmt')
ELSE_BLOCK = declare('else_block')

# While statement
# ---------------
WHILE_STMT = declare('while_stmt')

# For statement
# -------------
FOR_STMT = declare('for_stmt')

# With statement
# --------------
WITH_STMT = declare('with_stmt')
WITH_ITEM = declare('with_item')

# Try statement
# -------------
TRY_STMT = declare('try_stmt')


# Except statement
# ----------------
EXCEPT_BLOCK = declare('except_block')
EXCEPT_STAR_BLOCK = declare('except_star_block')
FINALLY_BLOCK = declare('finally_block')

# Match statement
# ---------------
MATCH_STMT = declare('match_stmt')
SUBJECT_EXPR = declare('subject_expr')
CASE_BLOCK = declare('case_block')
GUARD = declare('guard')
PATTERNS = declare('patterns')
PATTERN = declare('pattern')
AS_PATTERN = declare('as_pattern')
OR_PATTERN = declare('or_pattern')
CLOSED_PATTERN = declare('closed_pattern')
LITERAL_PATTERN = declare('literal_pattern')
LITERAL_EXPR = declare('literal_expr')
COMPLEX_NUMBER = declare('complex_number')
SIGNED_NUMBER = declare('signed_number')
SIGNED_REAL_NUMBER = declare('signed_real_number')
REAL_NUMBER = declare('real_number')
IMAGINARY_NUMBER = declare('imaginary_number')
CAPTURE_PATTERN = declare('capture_pattern')
PATTERN_CAPTURE_TARGET = declare('pattern_capture_target')
WILDCARD_PATTERN = declare('wildcard_pattern')
VALUE_PATTERN = declare('value_pattern')
ATTR = declare('attr')
NAME_OR_ATTR = declare('name_or_attr')
GROUP_PATTERN = declare('group_pattern')
SEQUENCE_PATTERN = declare('sequence_pattern')
OPEN_SEQUENCE_PATTERN = declare('open_sequence_pattern')
MAYBE_SEQUENCE_PATTERN = declare('maybe_sequence_pattern')
MAYBE_STAR_PATTERN = declare('maybe_star_pattern')
STAR_PATTERN = declare('star_pattern')
MAPPING_PATTERN = declare('mapping_pattern')
ITEMS_PATTERN = declare('items_pattern')
KEY_VALUE_PATTERN = declare('key_value_pattern')
DOUBLE_STAR_PATTERN = declare('double_star_pattern')
CLASS_PATTERN = declare('class_pattern')
POSITIONAL_PATTERNS = declare('positional_patterns')
KEYWORD_PATTERNS = declare('keyword_patterns')
KEYWORD_PATTERN = declare('keyword_pattern')

# Type statement
# ---------------
TYPE_ALIAS = declare('type_alias')

# Type parameter declaration
# --------------------------
TYPE_PARAMS = declare('type_params')
TYPE_PARAM_SEQ = declare('type_param_seq')
TYPE_PARAM = declare('type_param')
TYPE_PARAM_BOUND = declare('type_param_bound')
TYPE_PARAM_DEFAULT = declare('type_param_default')
TYPE_PARAM_STARRED_DEFAULT = declare('type_param_starred_default')

# EXPRESSIONS
# -----------
EXPRESSIONS = declare('expressions')
EXPRESSION = declare('expression')
YIELD_EXPR = declare('yield_expr')
STAR_EXPRESSIONS = declare('star_expressions')
STAR_EXPRESSION = declare('star_expression')
STAR_NAMED_EXPRESSIONS = declare('star_named_expressions')
STAR_NAMED_EXPRESSION = declare('star_named_expression')
ASSIGNMENT_EXPRESSION = declare('assignment_expression')
NAMED_EXPRESSION = declare('named_expression')
DISJUNCTION = declare('disjunction')
CONJUNCTION = declare('conjunction')
INVERSION = declare('inversion')

# Comparison operators
# --------------------
COMPARISON = declare('comparison')
COMPARE_OP_BITWISE_OR_PAIR = declare('compare_op_bitwise_or_pair')
EQ_BITWISE_OR = declare('eq_bitwise_or')
NOTEQ_BITWISE_OR = declare('noteq_bitwise_or')
LTE_BITWISE_OR = declare('lte_bitwise_or')
LT_BITWISE_OR = declare('lt_bitwise_or')
GTE_BITWISE_OR = declare('gte_bitwise_or')
GT_BITWISE_OR = declare('gt_bitwise_or')
NOTIN_BITWISE_OR = declare('notin_bitwise_or')
IN_BITWISE_OR = declare('in_bitwise_or')
ISNOT_BITWISE_OR = declare('isnot_bitwise_or')
IS_BITWISE_OR = declare('is_bitwise_or')

# Bitwise operators
# -----------------
BITWISE_OR = declare('bitwise_or')
BITWISE_XOR = declare('bitwise_xor')
BITWISE_AND = declare('bitwise_and')
SHIFT_EXPR = declare('shift_expr')

# Arithmetic operators
# --------------------
SUM = declare('sum')
TERM = declare('term')
FACTOR = declare('factor')
POWER = declare('power')

# Primary elements
# ----------------
AWAIT_PRIMARY = declare('await_primary')
PRIMARY = declare('primary')
SLICES = declare('slices')
SLICE = declare('slice')
ATOM = declare('atom')
GROUP = declare('group')

# Lambda functions
# ----------------
LAMBDEF = declare('lambdef')
LAMBDA_PARAMS = declare('lambda_params')
LAMBDA_PARAMETERS = declare('lambda_parameters')
LAMBDA_SLASH_NO_DEFAULT = declare('lambda_slash_no_default')
LAMBDA_SLASH_WITH_DEFAULT = declare('lambda_slash_with_default')
LAMBDA_STAR_ETC = declare('lambda_star_etc')
LAMBDA_KWDS = declare('lambda_kwds')
LAMBDA_PARAM_NO_DEFAULT = declare('lambda_param_no_default')
LAMBDA_PARAM_WITH_DEFAULT = declare('lambda_param_with_default')
LAMBDA_PARAM_MAYBE_DEFAULT = declare('lambda_param_maybe_default')

# LITERALS
# ========
FSTRING_MIDDLE = declare('fstring_middle')
FSTRING_REPLACEMENT_FIELD = declare('fstring_replacement_field')
FSTRING_CONVERSION = declare('fstring_conversion')
FSTRING_FULL_FORMAT_SPEC = declare('fstring_full_format_spec')
FSTRING_FORMAT_SPEC = declare('fstring_format_spec')
FSTRING = declare('fstring')
TSTRING_FORMAT_SPEC_REPLACEMENT_FIELD = declare('tstring_format_spec_replacement_field')
TSTRING_FORMAT_SPEC = declare('tstring_format_spec')
TSTRING_FULL_FORMAT_SPEC = declare('tstring_full_format_spec')
TSTRING_REPLACEMENT_FIELD = declare('tstring_replacement_field')
TSTRING_MIDDLE = declare('tstring_middle')
TSTRING = declare('tstring')
STRING = declare('string')
STRINGS = declare('strings')
LIST = declare('list')
TUPLE = declare('tuple')
SET = declare('set')
# Dicts
# -----
DICT = declare('dict')
DOUBLE_STARRED_KVPAIRS = declare('double_starred_kvpairs')
DOUBLE_STARRED_KVPAIR = declare('double_starred_kvpair')
KVPAIR = declare('kvpair')

# Comprehensions & Generators
# ---------------------------
FOR_IF_CLAUSES = declare('for_if_clauses')
FOR_IF_CLAUSE = declare('for_if_clause')
LISTCOMP = declare('listcomp')
SETCOMP = declare('setcomp')
GENEXP = declare('genexp')
DICTCOMP = declare('dictcomp')

# FUNCTION CALL ARGUMENTS
# =======================
ARGUMENTS = declare('arguments')
ARGS = declare('args')
KWARGS = declare('kwargs')
STARRED_EXPRESSION = declare('starred_expression')
KWARG_OR_STARRED = declare('kwarg_or_starred')
KWARG_OR_DOUBLE_STARRED = declare('kwarg_or_double_starred')

# ASSIGNMENT TARGETS
# ==================

# Generic targets
# ---------------
STAR_TARGETS = declare('star_targets')
STAR_TARGETS_LIST_SEQ = declare('star_targets_list_seq')
STAR_TARGETS_TUPLE_SEQ = declare('star_targets_tuple_seq')
STAR_TARGET = declare('star_target')
TARGET_WITH_STAR_ATOM = declare('target_with_star_atom')
STAR_ATOM = declare('star_atom')
SINGLE_TARGET = declare('single_target')
SINGLE_SUBSCRIPT_ATTRIBUTE_TARGET = declare('single_subscript_attribute_target')
T_PRIMARY = declare('t_primary')
T_LOOKAHEAD = declare('t_lookahead')

# Targets for del statements
# --------------------------
DEL_TARGETS = declare('del_targets')
DEL_TARGET = declare('del_target')
DEL_T_ATOM = declare('del_t_atom')

# TYPING ELEMENTS
# ---------------
TYPE_EXPRESSIONS = declare('type_expressions')
FUNC_TYPE_COMMENT = declare('func_type_comment')

# INVALID RULES
# =============
INVALID_ARGUMENTS = declare('invalid_arguments')
EXPRESSION_WITHOUT_INVALID = declare('expression_without_invalid')
INVALID_LEGACY_EXPRESSION = declare('invalid_legacy_expression')
INVALID_TYPE_PARAM = declare('invalid_type_param')
INVALID_EXPRESSION = declare('invalid_expression')
INVALID_NAMED_EXPRESSION = declare('invalid_named_expression')
INVALID_ASSIGNMENT = declare('invalid_assignment')
INVALID_ANN_ASSIGN_TARGET = declare('invalid_ann_assign_target')
INVALID_RAISE_STMT = declare('invalid_raise_stmt')
INVALID_DEL_STMT = declare('invalid_del_stmt')
INVALID_BLOCK = declare('invalid_block')
INVALID_COMPREHENSION = declare('invalid_comprehension')
INVALID_DICT_COMPREHENSION = declare('invalid_dict_comprehension')
INVALID_PARAMETERS = declare('invalid_parameters')
INVALID_DEFAULT = declare('invalid_default')
INVALID_STAR_ETC = declare('invalid_star_etc')
INVALID_KWDS = declare('invalid_kwds')
INVALID_PARAMETERS_HELPER = declare('invalid_parameters_helper')
INVALID_LAMBDA_PARAMETERS = declare('invalid_lambda_parameters')
INVALID_LAMBDA_PARAMETERS_HELPER = declare('invalid_lambda_parameters_helper')
INVALID_LAMBDA_STAR_ETC = declare('invalid_lambda_star_etc')
INVALID_LAMBDA_KWDS = declare('invalid_lambda_kwds')
INVALID_DOUBLE_TYPE_COMMENTS = declare('invalid_double_type_comments')
INVALID_WITH_ITEM = declare('invalid_with_item')
INVALID_FOR_IF_CLAUSE = declare('invalid_for_if_clause')
INVALID_FOR_TARGET = declare('invalid_for_target')
INVALID_GROUP = declare('invalid_group')
INVALID_IMPORT = declare('invalid_import')
INVALID_DOTTED_AS_NAME = declare('invalid_dotted_as_name')
INVALID_IMPORT_FROM_AS_NAME = declare('invalid_import_from_as_name')
INVALID_IMPORT_FROM_TARGETS = declare('invalid_import_from_targets')
INVALID_WITH_STMT = declare('invalid_with_stmt')
INVALID_WITH_STMT_INDENT = declare('invalid_with_stmt_indent')
INVALID_TRY_STMT = declare('invalid_try_stmt')
INVALID_EXCEPT_STMT = declare('invalid_except_stmt')
INVALID_EXCEPT_STAR_STMT = declare('invalid_except_star_stmt')
INVALID_FINALLY_STMT = declare('invalid_finally_stmt')
INVALID_EXCEPT_STMT_INDENT = declare('invalid_except_stmt_indent')
INVALID_EXCEPT_STAR_STMT_INDENT = declare('invalid_except_star_stmt_indent')
INVALID_MATCH_STMT = declare('invalid_match_stmt')
INVALID_CASE_BLOCK = declare('invalid_case_block')
INVALID_AS_PATTERN = declare('invalid_as_pattern')
INVALID_CLASS_PATTERN = declare('invalid_class_pattern')
INVALID_CLASS_ARGUMENT_PATTERN = declare('invalid_class_argument_pattern')
INVALID_IF_STMT = declare('invalid_if_stmt')
INVALID_ELIF_STMT = declare('invalid_elif_stmt')
INVALID_ELSE_STMT = declare('invalid_else_stmt')
INVALID_WHILE_STMT = declare('invalid_while_stmt')
INVALID_FOR_STMT = declare('invalid_for_stmt')
INVALID_DEF_RAW = declare('invalid_def_raw')
INVALID_CLASS_DEF_RAW = declare('invalid_class_def_raw')
INVALID_DOUBLE_STARRED_KVPAIRS = declare('invalid_double_starred_kvpairs')
INVALID_KVPAIR = declare('invalid_kvpair')
INVALID_STARRED_EXPRESSION_UNPACKING = declare('invalid_starred_expression_unpacking')
INVALID_STARRED_EXPRESSION = declare('invalid_starred_expression')
INVALID_FSTRING_REPLACEMENT_FIELD = declare('invalid_fstring_replacement_field')
INVALID_FSTRING_CONVERSION_CHARACTER = declare('invalid_fstring_conversion_character')
INVALID_TSTRING_REPLACEMENT_FIELD = declare('invalid_tstring_replacement_field')
INVALID_TSTRING_CONVERSION_CHARACTER = declare('invalid_tstring_conversion_character')
INVALID_ARITHMETIC = declare('invalid_arithmetic')
INVALID_FACTOR = declare('invalid_factor')
INVALID_TYPE_PARAMS = declare('invalid_type_params')



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception


from tapl_lang.core import chunker
from tapl_lang.dev import chunk_dumper


def check(text: str, expected: str) -> None:
    chunks = chunker.chunk_text(text.strip())
    dump = chunk_dumper.get_dump(chunks)
    assert dump.strip() == expected.strip()


def test_decode_chunk1():
    check(
        """
Hello
World
""",
        """
    0:chunk line_records_length=1 children_length=0
1:0  |Hello\\n
    0:chunk line_records_length=1 children_length=0
2:0  |World
""",
    )


def test_decode_chunk2():
    check(
        """
Hello
 World
""",
        """
    0:chunk line_records_length=2 children_length=0
1:0  |Hello\\n
2:1  | World
""",
    )


def test_decode_chunk3():
    check(
        """
Hello
 World
One
    two
  three
""",
        """
    0:chunk line_records_length=2 children_length=0
1:0  |Hello\\n
2:1  | World\\n
    0:chunk line_records_length=3 children_length=0
3:0  |One\\n
4:4  |    two\\n
5:2  |  three
""",
    )


def test_decode_chunk4():
    check(
        """
Hello:
 World
""",
        """
    0:chunk line_records_length=1 children_length=1
1:0  |Hello:\\n
    1: chunk line_records_length=1 children_length=0
2:1  | World
""",
    )


def test_decode_chunk5():
    check(
        """
def compute_next(n):
  if n % 2 == 0:
    print('even')
    return n / 2
  else:
    print('odd')
    return 3 * n + 1
""",
        """
    0:chunk line_records_length=1 children_length=2
1:0  |def compute_next(n):\\n
    2:  chunk line_records_length=1 children_length=2
2:2  |  if n % 2 == 0:\\n
    4:    chunk line_records_length=1 children_length=0
3:4  |    print('even')\\n
    4:    chunk line_records_length=1 children_length=0
4:4  |    return n / 2\\n
    2:  chunk line_records_length=1 children_length=2
5:2  |  else:\\n
    4:    chunk line_records_length=1 children_length=0
6:4  |    print('odd')\\n
    4:    chunk line_records_length=1 children_length=0
7:4  |    return 3 * n + 1
""",
    )



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from __future__ import annotations

import dataclasses
import enum
import io
import itertools
import logging
from collections.abc import Callable, Iterable

from tapl_lang.core import line_record, syntax, tapl_error
from tapl_lang.lib import terms

logger = logging.getLogger(__name__)

# Implemented PEG parser - https://en.wikipedia.org/wiki/Parsing_expression_grammar,
# https://pdos.csail.mit.edu/~baford/packrat/thesis/
# Left recursion: https://web.cs.ucla.edu/~todd/research/pepm08.pdf
# Error Detection taken from - https://arxiv.org/abs/1806.11150


ParseFunction = Callable[['Cursor'], syntax.Term]
OrderedParseFunctions = Iterable[ParseFunction | str]
GrammarRuleMap = dict[str, OrderedParseFunctions]


@dataclasses.dataclass(frozen=True)
class Grammar:
    rule_map: GrammarRuleMap
    start_rule: str

    def clone(self) -> Grammar:
        return Grammar(self.rule_map.copy(), self.start_rule)


def parse_function_name(function: ParseFunction | str) -> str:
    if isinstance(function, str):
        return f'|>{function}'
    return f'@{function.__name__}'


@dataclasses.dataclass
class Config:
    mode: syntax.Term


class Cursor:
    def __init__(self, row: int, col: int, config: Config, engine: PegEngine) -> None:
        self.row = row
        self.col = col
        self.config = config
        self.engine = engine

    def clone(self) -> Cursor:
        return Cursor(self.row, self.col, self.config, self.engine)

    def copy_position_from(self, other: Cursor) -> None:
        if self.engine is not other.engine:
            raise tapl_error.TaplError('Both cursors do not have a same engine instance.')
        self.row = other.row
        self.col = other.col

    def assert_position(self) -> None:
        if not (0 <= self.row < len(self.engine.line_records)):
            raise tapl_error.TaplError('Cursor row is out of range.')
        if not (0 <= self.col < len(self.engine.line_records[self.row].text)):
            raise tapl_error.TaplError('Cursor col is out of range.')

    def current_char(self) -> str:
        self.assert_position()
        return self.engine.line_records[self.row].text[self.col]

    def is_end(self) -> bool:
        if self.row == len(self.engine.line_records):
            if self.col != 0:
                raise tapl_error.TaplError('When cursor ends, col must be 0.')
            return True
        self.assert_position()
        return False

    def move_to_next(self) -> bool:
        if self.is_end():
            return False
        self.col += 1
        if self.col == len(self.engine.line_records[self.row].text):
            self.row += 1
            self.col = 0
        return True

    def current_position(self) -> syntax.Position:
        if self.is_end():
            line_record = self.engine.line_records[self.row - 1]
            return syntax.Position(line_record.line_number, len(line_record.text))
        self.assert_position()
        return syntax.Position(self.engine.line_records[self.row].line_number, self.col)

    def consume_rule(self, rule: str, config: Config | None = None) -> syntax.Term:
        term, self.row, self.col = self.engine.apply_rule(self.row, self.col, rule, config or self.config)
        return term

    def start_tracker(self) -> Tracker:
        return Tracker(self)

    def skip_whitespace(self) -> None:
        while not self.is_end() and self.current_char().isspace():
            self.move_to_next()

    def consume_text(self, text: str) -> bool:
        for char in text:
            if self.is_end() or self.current_char() != char:
                return False
            self.move_to_next()
        return True


ParseFailed = syntax.ErrorTerm(message='Parsing failed: Unable to match any rule.')


# Tracker is designed for use within its originating function only and should not be passed between functions.
# Cursor, on the other hand, can be passed freely between functions.
class Tracker:
    def __init__(self, cursor: Cursor) -> None:
        self.cursor = cursor
        self.start_position = cursor.current_position()
        self.captured_error: syntax.ErrorTerm | None = None

    @property
    def location(self):
        return syntax.Location(start=self.start_position, end=self.cursor.current_position())

    def fail(self):
        return self.captured_error or ParseFailed

    def validate(self, term: syntax.Term) -> bool:
        if self.captured_error:
            return False
        if term is ParseFailed:
            return False
        if isinstance(term, syntax.ErrorTerm):
            self.captured_error = term
            return False
        return term is not None


class CellState(enum.IntEnum):
    BLANK = 1
    START = 2
    DONE = 3


@dataclasses.dataclass(frozen=True)
class CellKey:
    row: int
    col: int
    rule: str


@dataclasses.dataclass
class Cell:
    next_row: int
    next_col: int
    growable: bool
    state: CellState
    term: syntax.Term


CellMemo = dict[CellKey, Cell]


class PegEngine:
    def __init__(self, line_records: list[line_record.LineRecord], grammar_rule_map: GrammarRuleMap):
        self.line_records = line_records
        self.grammar_rule_map = grammar_rule_map
        self.cell_memo: CellMemo = {}
        # Set a call stack limit to prevent infinite recursion in rule applications
        self.rule_call_stack_limit = 1000

    def call_parse_function(
        self, key: CellKey, function: ParseFunction | str, config: Config
    ) -> tuple[syntax.Term, int, int]:
        cursor = Cursor(key.row, key.col, config=config, engine=self)

        def create_location() -> syntax.Location:
            start_cursor = Cursor(key.row, key.col, config=config, engine=self)
            return syntax.Location(start=start_cursor.current_position(), end=cursor.current_position())

        def route(rule: str) -> ParseFunction:
            def parse(c: Cursor) -> syntax.Term:
                return c.consume_rule(rule)

            return parse

        try:
            if isinstance(function, str):
                term = route(function)(cursor)
            else:
                term = function(cursor)
            if term is None:
                term = syntax.ErrorTerm(
                    message=f'PegEngine: rule={key.rule}:{parse_function_name(function)} returned None.',
                    location=create_location(),
                )
        except Exception as e:  # noqa: BLE001  The user provided function may raise any exception.
            term = syntax.ErrorTerm(
                message=f'PegEngine: rule={key.rule}:{parse_function_name(function)} error={e}',
                location=create_location(),
            )
        return term, cursor.row, cursor.col

    def call_ordered_parse_functions(self, key: CellKey, config: Config) -> tuple[syntax.Term, int, int]:
        functions = self.grammar_rule_map.get(key.rule)
        if not functions:
            raise tapl_error.TaplError(f'Rule "{key.rule}" is not defined in the Grammar.')
        for fn in functions:
            term, row, col = self.call_parse_function(key, fn, config=config)
            if term is not ParseFailed:
                return term, row, col
        return ParseFailed, key.row, key.col

    def grow_seed(self, key: CellKey, cell: Cell, config: Config) -> None:
        seed_next_row, seed_next_col = cell.next_row, cell.next_col
        iteration_count = 10  # Prevent infinite loop by limiting iterations
        while iteration_count > 0:
            iteration_count -= 1
            term, next_row, next_col = self.call_ordered_parse_functions(key, config)
            if term is ParseFailed:
                cell.term = syntax.ErrorTerm(
                    message='PegEngine: Once ordered_parse_functions was successful, but it failed afterward. This indicates an inconsistency between ordered parse functions.'
                )
                return
            if isinstance(term, syntax.ErrorTerm):
                cell.term = term
                return
            # Stop growing when the new next position mathches seed's next position, as this indicates a cycle.
            if next_row == seed_next_row and next_col == seed_next_col:
                return
            cell.term, cell.next_row, cell.next_col = term, next_row, next_col
        cell.term = syntax.ErrorTerm(message='PegEngine: Growing failed due to too many iterations.')

    def apply_rule(self, row: int, col: int, rule: str, config: Config) -> tuple[syntax.Term, int, int]:
        self.rule_call_stack_limit -= 1
        if self.rule_call_stack_limit < 0:
            error = syntax.ErrorTerm(message='PEG Parser: Rule application limit exceeded.')
            return (error, row, col)
        cell_key = CellKey(row, col, rule)
        cell = self.cell_memo.get(cell_key)
        if not cell:
            cell = Cell(
                next_row=cell_key.row, next_col=cell_key.col, growable=False, state=CellState.BLANK, term=ParseFailed
            )
            self.cell_memo[cell_key] = cell
        match cell.state:
            case CellState.BLANK:
                cell.state = CellState.START
                cell.term, cell.next_row, cell.next_col = self.call_ordered_parse_functions(cell_key, config)
                cell.state = CellState.DONE
                if cell.growable and not isinstance(cell.term, syntax.ErrorTerm):
                    self.grow_seed(cell_key, cell, config)
            case CellState.START:
                # Left recursion detected. Delaying expansion of this rule.
                cell.growable = True
            case CellState.DONE:
                # Rule already parsed at this position, so no further action is required.
                pass
            case _:
                cell.term = syntax.ErrorTerm(f'PEG Parser Engine: Unknown cell state [{cell.state}] at {cell_key}.')
        self.rule_call_stack_limit += 1
        return cell.term, cell.next_row, cell.next_col

    def dump(self) -> str:
        return 'Use PegEngineDebug to get the engine dump.'


@dataclasses.dataclass
class ParseTrace:
    start_row: int
    start_col: int
    end_row: int
    end_col: int
    rule: str
    function_name: str
    term: syntax.Term
    start_call_order: int
    end_call_order: int
    growing_id: int | None
    applyied_rules: list[str]


class PegEngineDebug(PegEngine):
    def __init__(self, line_records: list[line_record.LineRecord], grammar_rule_map: GrammarRuleMap):
        super().__init__(line_records, grammar_rule_map)
        self.parse_traces: list[ParseTrace] = []
        self._next_call_order = 0
        self.growing_id: int | None = None
        self.next_growing_id = 0
        self.applied_rules: list[str] = []

    def next_call_order(self) -> int:
        self._next_call_order += 1
        return self._next_call_order

    def call_parse_function(
        self, key: CellKey, function: ParseFunction | str, config: Config
    ) -> tuple[syntax.Term, int, int]:
        old_applied_rules = self.applied_rules
        self.applied_rules = []
        start_call_order = self.next_call_order()
        term, row, col = super().call_parse_function(key, function, config=config)
        self.parse_traces.append(
            ParseTrace(
                start_row=key.row,
                start_col=key.col,
                end_row=row,
                end_col=col,
                rule=key.rule,
                function_name=parse_function_name(function),
                term=term,
                start_call_order=start_call_order,
                end_call_order=self.next_call_order(),
                growing_id=self.growing_id,
                applyied_rules=self.applied_rules,
            )
        )
        self.applied_rules = old_applied_rules
        return term, row, col

    def grow_seed(self, key: CellKey, cell: Cell, config: Config) -> None:
        old_growing_id = self.growing_id
        self.next_growing_id += 1
        self.growing_id = self.next_growing_id
        super().grow_seed(key, cell, config)
        self.growing_id = old_growing_id

    def apply_rule(self, row: int, col: int, rule: str, config: Config) -> tuple[syntax.Term, int, int]:
        self.applied_rules.append(f'{row}:{col}:{rule}')
        term, next_row, next_col = super().apply_rule(row, col, rule, config)
        if self.cell_memo[CellKey(row, col, rule)].state == CellState.START:
            self.applied_rules[-1] += ' (left recursion)'
        return term, next_row, next_col

    def dump_term(self, term: syntax.Term | None) -> tuple[str, str]:
        if term is None:
            return 'Error', 'None'
        if term is ParseFailed:
            return 'Fail', ''
        if isinstance(term, syntax.ErrorTerm):
            return 'Error', term.message
        return 'Success', f'term={term.__class__.__qualname__}'

    def dump_table(self, output: io.StringIO, table: list[list[str]]) -> None:
        col_widths = [max(len(str(item)) for item in col) for col in zip(*table, strict=True)]
        row_format = ' '.join(f'{{:<{width + 2}}}' for width in col_widths[:-1]) + ' {}'
        output.writelines(row_format.format(*row) + '\n' for row in table)

    def dump_cell_memo(self) -> list[list[str]]:
        table = [['Start/Rule', 'End', 'Status', 'Details']]
        sorted_cells = sorted(self.cell_memo.items(), key=lambda item: (item[0].row, item[0].col, item[0].rule))
        for (row, col), group in itertools.groupby(sorted_cells, key=lambda item: (item[0].row, item[0].col)):
            table.append([f'{row}:{col}', '', '', ''])
            for item in group:
                cell = item[1]
                if cell.state == CellState.DONE:
                    state, details = self.dump_term(cell.term)
                else:
                    state = cell.state.name.capitalize()
                    details = ''
                state += ' Grown' if cell.growable else ''
                table.append([f'   {item[0].rule}', f'{cell.next_row}:{cell.next_col}', state, details])
        return table

    def tableize_parse_traces(self) -> list[list[str]]:
        sorted_traces = sorted(self.parse_traces, key=lambda t: (t.start_row, t.start_col, t.rule, t.start_call_order))
        table = [['Start/Rule', 'End', 'Order#', 'Status/Grow', 'Applied rules', 'Details']]
        last_pos = None
        for (row, col, rule), group in itertools.groupby(
            sorted_traces, key=lambda t: (t.start_row, t.start_col, t.rule)
        ):
            pos = f'{row}:{col}'
            if last_pos != pos:
                table.append(['', '', '', '', '', ''])
            last_pos = pos
            table.append([f'{row}:{col}:{rule}', '', '', '', '', ''])
            for trace in group:
                rule_key = f'   {trace.function_name}'
                status, details = self.dump_term(trace.term)
                if trace.growing_id:
                    status += ' G' + str(trace.growing_id)
                table.append(
                    [
                        rule_key,
                        f'{trace.end_row}:{trace.end_col}',
                        f'{trace.start_call_order}..{trace.end_call_order}',
                        status,
                        ', '.join(trace.applyied_rules),
                        details,
                    ]
                )
        return table

    def dump(self) -> str:
        output = io.StringIO()
        output.write('\n------PEG Engine Dump (Rows sorted by row,col,rule,call_order)--------')
        # Print line records
        max_len = max(len(line.text) for line in self.line_records)
        output.write('\n  ')
        for i in range(max_len):
            output.write(str(i % 10))
        output.write('\n')
        for i, line in enumerate(self.line_records):
            output.write(f'{i % 10}|')
            output.write(line.text)

        output.write('\n\n')
        self.dump_table(output, self.dump_cell_memo())
        output.write('\n\n')
        self.dump_table(output, self.tableize_parse_traces())
        output.write('\n------End of PEG Engine Dump------------------------------------------\n')
        return output.getvalue()


def find_first_position(line_records: list[line_record.LineRecord]) -> tuple[int, int]:
    for row in range(len(line_records)):
        for col in range(len(line_records[row].text)):
            return row, col
    return len(line_records), 0


def parse_line_records(
    line_records: list[line_record.LineRecord], grammar: Grammar, *, debug: bool = False, config: Config | None = None
) -> syntax.Term:
    config = config or Config(mode=terms.MODE_SAFE)
    engine = PegEngineDebug(line_records, grammar.rule_map) if debug else PegEngine(line_records, grammar.rule_map)
    row, col = find_first_position(line_records)
    if row == len(line_records) and col == 0:
        return syntax.ErrorTerm(message='Empty text.')
    term, next_row, next_col = engine.apply_rule(row, col, grammar.start_rule, config=config)
    if debug:
        logger.warning(engine.dump())
    if not isinstance(term, syntax.ErrorTerm) and not (next_row == len(line_records) and next_col == 0):
        lineno = line_records[0].line_number if line_records else -1
        return syntax.ErrorTerm(
            message=f'chunk[line:{lineno}] Not all text consumed: indices {next_row}:{next_col}/{len(line_records)}:0.',
        )
    return term


def parse_text(text: str, grammar: Grammar, *, debug: bool = False, config: Config | None = None) -> syntax.Term:
    return parse_line_records(line_record.split_text_to_lines(text), grammar, debug=debug, config=config)



from tapl_lang.pythonlike.predef1 import predef_scope as predef_scope__sa
s0 = predef_scope__sa.tapl_typing.create_scope(parent__sa=predef_scope__sa)

class Vector_:

    def __init__(self, n):
        s1 = s0.tapl_typing.create_scope(parent__sa=s0, self=self, n=n)
        s1.self.n = s1.n
        return s1.tapl_typing.get_return_type(s1)

    def concat(self, m):
        s1 = s0.tapl_typing.create_scope(parent__sa=s0, self=self, m=m)
        s1.tapl_typing.add_return_type(s1, s1.Vector(s1.n + s1.m))
        return s1.tapl_typing.get_return_type(s1)
s0.Vector, s0.Vector_ = s0.tapl_typing.create_class(cls=Vector_, init_args=[s0.Int], methods=[('concat', [s0.Int])])
s0.print(s0.Str)


from tapl_lang.lib import tapl_typing
from tapl_lang.pythonlike.predef1 import predef_scope as predef_scope__sa
s0 = tapl_typing.create_scope(parent__sa=predef_scope__sa)
tapl_typing.import_module(s0, ['tapl_dev'], 'tapl_lang.lib', 0)
tapl_typing.import_module(s0, ['examples.easy'])
s0.tapl_dev.log(s0.examples.easy.factorial(s0.Int))
s0.my_dog = s0.examples.easy.Dog(s0.Str)
s0.print(s0.my_dog.bark())
s0.print(s0.examples.easy.greet_dog(s0.my_dog))
s0.examples.easy.main()



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from typing import Any

OP_LABEL = {
    '__add__': '+',
    '__sub__': '-',
    '__mul__': '*',
    '__truediv__': '/',
    '__floordiv__': '//',
    '__mod__': '%',
    '__pow__': '**',
    '__lshift__': '<<',
    '__rshift__': '>>',
    '__or__': '|',
    '__xor__': '^',
    '__and__': '&',
    '__matmul__': '@',
    # Comparison operators
    '__eq__': '==',
    '__ne__': '!=',
    '__lt__': '<',
    '__le__': '<=',
    '__gt__': '>',
    '__ge__': '>=',
}

# The __sa suffix (static attribute) indicates fields that are skipped by the DynamicAttributeMixin


# ruff: noqa: N805
class DynamicAttributeMixin:
    """Proxy mixin for dynamic attribute access."""

    def get_label__sa(self) -> str:
        return f'{self.__class__.__name__} class'

    def load__sa(self, key: str) -> Any:
        raise NotImplementedError(
            f'{self.__class__.__name__} class does not support loading attribute: {self.get_label__sa()}.{key}'
        )

    def store__sa(self, key: str, value: Any) -> None:
        del value  # unused
        raise NotImplementedError(
            f'{self.__class__.__name__} class does not support storing attribute: {self.get_label__sa()}.{key}'
        )

    def delete__sa(self, key: str) -> None:
        raise NotImplementedError(
            f'{self.__class__.__name__} class does not support deleting attribute: {self.get_label__sa()}.{key}'
        )

    def call_binop__sa(left, op: str, right: Any) -> Any | None:
        try:
            return left.load__sa(op)(right)
        except Exception as e:
            label = OP_LABEL.get(op, op)
            raise TypeError(f'unsupported operand type(s) for {label}: {left} and {right}. {e}') from e

    def __getattr__(self__sa, name):
        return self__sa.load__sa(name)

    def __setattr__(self__sa, name: str, value: Any):
        if name.endswith('__sa'):
            super().__setattr__(name, value)
        else:
            self__sa.store__sa(name, value)

    def __delattr__(self__sa, name: str):
        if name.endswith('__sa'):
            super().__delattr__(name)
        else:
            self__sa.delete__sa(name)

    def __call__(self__sa, *args, **kwargs):
        return self__sa.load__sa('__call__')(*args, **kwargs)

    # Numeric methods
    def __trunc__(self__sa):
        return self__sa.load__sa('__trunc__')()

    def __floor__(self__sa):
        return self__sa.load__sa('__floor__')()

    def __ceil__(self__sa):
        return self__sa.load__sa('__ceil__')()

    def __round__(self__sa, n: int = 0):
        return self__sa.load__sa('__round__')(n)

    def __invert__(self__sa):
        return self__sa.load__sa('__invert__')()

    def __neg__(self__sa):
        return self__sa.load__sa('__neg__')()

    def __pos__(self__sa):
        return self__sa.load__sa('__pos__')()

    def __abs__(self__sa):
        return self__sa.load__sa('__abs__')()

    def __int__(self__sa):
        return self__sa.load__sa('__int__')()

    def __float__(self__sa):
        return self__sa.load__sa('__float__')()

    def __complex__(self__sa):
        return self__sa.load__sa('__complex__')()

    def __index__(self__sa):
        return self__sa.load__sa('__index__')()

    # Arithmetic operators

    def __add__(self__sa, other):
        return self__sa.call_binop__sa('__add__', other)

    def __sub__(self__sa, other):
        return self__sa.call_binop__sa('__sub__', other)

    def __mul__(self__sa, other):
        return self__sa.call_binop__sa('__mul__', other)

    def __truediv__(self__sa, other):
        return self__sa.call_binop__sa('__truediv__', other)

    def __floordiv__(self__sa, other):
        return self__sa.call_binop__sa('__floordiv__', other)

    def __mod__(self__sa, other):
        return self__sa.call_binop__sa('__mod__', other)

    def __divmod__(self__sa, other):
        return self__sa.load__sa('__divmod__')(other)

    def __pow__(self__sa, other):
        return self__sa.call_binop__sa('__pow__', other)

    def __lshift__(self__sa, other):
        return self__sa.call_binop__sa('__lshift__', other)

    def __rshift__(self__sa, other):
        return self__sa.call_binop__sa('__rshift__', other)

    def __and__(self__sa, other):
        return self__sa.call_binop__sa('__and__', other)

    def __or__(self__sa, other):
        return self__sa.call_binop__sa('__or__', other)

    def __xor__(self__sa, other):
        return self__sa.call_binop__sa('__xor__', other)

    def __matmul__(self__sa, other):
        return self__sa.call_binop__sa('__matmul__', other)

    # String methods

    # FIXME: Python __str__ method has special handling
    # def __str__(self__sa):
    #     return self__sa.load__sa('__str__')()

    def __repr__(self__sa):
        return self__sa.load__sa('__repr__')()

    def __unicode__(self__sa):
        return self__sa.load__sa('__unicode__')()

    # FIXME: Python __format__ method has special handling
    # def __format__(self__sa, format_spec):
    #     return self__sa.load__sa('__format__')(format_spec)

    # Python __hash__ and __eq__ methods have special handling, so we manually dispatch them to __hash__sa and __eq__sa methods.
    # def __hash__(self__sa):
    #     return self__sa.load__sa('__hash__')()

    def __nonzero__(self__sa):
        return self__sa.load__sa('__nonzero__')()

    def __dir__(self__sa):
        return self__sa.load__sa('__dir__')()

    def __sizeof__(self__sa):
        return self__sa.load__sa('__sizeof__')()

    # Comparison operators

    # Python __hash__ and __eq__ methods have special handling, so we manually dispatch them to __hash__sa and __eq__sa methods.
    # def __eq__(self__sa, other):
    #     return self__sa.call_binop__sa('__eq__', other)  # operator: ==

    def __ne__(self__sa, other):
        return self__sa.call_binop__sa('__ne__', other)  # operator: !=

    def __lt__(self__sa, other):
        return self__sa.call_binop__sa('__lt__', other)  # operator: <

    def __le__(self__sa, other):
        return self__sa.call_binop__sa('__le__', other)  # operator: <=

    def __gt__(self__sa, other):
        return self__sa.call_binop__sa('__gt__', other)  # operator: >

    def __ge__(self__sa, other):
        return self__sa.call_binop__sa('__ge__', other)  # operator: >=

    # Subscript operators
    def __getitem__(self__sa, key):
        return self__sa.load__sa('__getitem__')(key)

    def __setitem__(self__sa, key, value):
        return self__sa.load__sa('__setitem__')(key, value)

    def __delitem__(self__sa, key):
        return self__sa.load__sa('__delitem__')(key)

    def __contains__(self__sa, item):
        return self__sa.load__sa('__contains__')(item)



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import ast
import io
import os
import pathlib
import subprocess

from approvaltests.approvals import verify
from approvaltests.core.namer import Namer

from tapl_lang.lib.compiler import compile_tapl


def run_command(command):
    """
    Runs a command and captures its stdout and stderr.

    Args:
      command: A string or list of strings representing the command to run.

    Returns:
      A tuple containing the return code, stdout, and stderr.
    """
    try:
        process = subprocess.run(
            command,
            capture_output=True,
            text=True,  # Decode stdout and stderr as text
            check=False,  # do not raise exception on non-zero return code.
        )
    except FileNotFoundError:
        return -1, '', f'Command not found: {command}'
    else:
        return process.returncode, process.stdout, process.stderr


class ApprovalNamer(Namer):
    def __init__(self, filename: str) -> None:
        self.filename = filename

    def get_received_filename(self, base: str | None = None) -> str:
        del base
        return self.filename + Namer.RECEIVED

    def get_approved_filename(self, base: str | None = None) -> str:
        del base
        return self.filename


def run_python_files(base_directory: str, filenames: list[str], output_file: str) -> None:
    output = io.StringIO()
    for filename in filenames:
        filepath = os.path.join(base_directory, filename)
        return_code, stdout, stderr = run_command(['python', filepath])
        output.write(f'====== {filename} exit-status:{return_code} stdout:\n')
        output.write(stdout)
        output.write('\n------ stderr:\n')
        output.write(stderr)
        output.write('\n------ end.\n\n')
        if return_code != 0:
            break
    filepath = os.path.join(base_directory, output_file)
    verify(output.getvalue(), namer=ApprovalNamer(filepath))


def run_golden_test(test_name: str, *, use_wrap: bool = False) -> None:
    base_directory = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'goldens')
    source = pathlib.Path(os.path.join(base_directory, f'{test_name}.tapl')).read_text()
    layers = compile_tapl(source)
    filenames = []

    for i in reversed(range(len(layers))):
        suffix = i if i else ''
        filename = f'{test_name}{suffix}.py'
        filenames.append(filename)
        filepath = os.path.join(base_directory, filename)
        verify(ast.unparse(layers[i]), namer=ApprovalNamer(filepath))
    if use_wrap:
        filenames = [f'{test_name}.wrap.py']
    run_python_files(base_directory, filenames, f'{test_name}.output')


def test_goldens():
    run_golden_test('grammar')
    run_golden_test('simple_function')
    run_golden_test('function_error', use_wrap=True)
    run_golden_test('simple_class')
    run_golden_test('importing')
    run_golden_test('sum')
    run_golden_test('collatz_conjecture')
    run_golden_test('type_constructions')
    run_golden_test('fibonacci')
    run_golden_test('try_stmt')
    run_golden_test('system_f')
    run_golden_test('matrix')



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception


import ast

from tapl_lang.core import syntax
from tapl_lang.core.chunker import chunk_text
from tapl_lang.core.parser import parse_text
from tapl_lang.lib import compiler, python_backend, scope, tapl_typing, terms
from tapl_lang.pythonlike import grammar, predef1
from tapl_lang.pythonlike.language import PythonlikeLanguage


def check_parsed_term(parsed: syntax.Term) -> None:
    if parsed is None:
        raise RuntimeError('Parser returns None.')
    error_bucket: list[syntax.ErrorTerm] = compiler.gather_errors(parsed)
    if error_bucket:
        messages = [e.message for e in error_bucket]
        raise SyntaxError('\n\n'.join(messages))


def parse_stmt(text: str, *, debug=False) -> list[ast.stmt]:
    parsed = parse_text(text, grammar.get_grammar(), debug=debug)
    placeholder = syntax.find_placeholder(parsed)
    if placeholder is not None:
        placeholder.is_placeholder = False
    check_parsed_term(parsed)
    safe_term = compiler.make_safe_term(parsed)
    layers = syntax.LayerSeparator(2).build(lambda layer: layer(safe_term))
    ast_generator = python_backend.AstGenerator()
    return [s for layer in layers for s in ast_generator.generate_stmt(layer, syntax.BackendSetting(scope_level=0))]


def run_stmt(stmts: list[ast.stmt]):
    compiled_code = compile(ast.Module(body=stmts, type_ignores=[]), filename='', mode='exec')
    globals_ = {'s0': scope.Scope(parent=predef1.predef_scope), 'tapl_typing': tapl_typing}
    return eval(compiled_code, globals_)


def parse_module(text: str) -> list[ast.AST]:
    chunks = chunk_text(text.strip())
    language = PythonlikeLanguage()
    module = terms.Module(body=[syntax.TermList(terms=[], is_placeholder=True)])
    language.parse_chunks(chunks, [module])
    check_parsed_term(module)
    ls = syntax.LayerSeparator(2)
    safe_module = compiler.make_safe_term(module)
    layers = ls.build(lambda layer: layer(safe_module))
    return [python_backend.AstGenerator().generate_ast(layer, syntax.BackendSetting(scope_level=0)) for layer in layers]


def test_assign_name():
    [stmt1, stmt2] = parse_stmt('a=1')
    assert ast.unparse(stmt1) == 'a = 1'
    assert ast.unparse(stmt2) == 's0.a = s0.Int'
    assert run_stmt([stmt2]) is None
    assert run_stmt([stmt1]) is None


def test_assign_empty_list():
    [stmt1, stmt2] = parse_stmt('a=[]')
    assert ast.unparse(stmt1) == 'a = []'
    assert ast.unparse(stmt2) == 's0.a = tapl_typing.create_typed_list()'
    assert run_stmt([stmt2]) is None
    assert run_stmt([stmt1]) is None


def test_assign_attribute():
    [stmt1, stmt2] = parse_stmt('a.b=1')
    assert ast.unparse(stmt1) == 'a.b = 1'
    assert ast.unparse(stmt2) == 's0.a.b = s0.Int'


def test_return1():
    [stmt1, stmt2] = parse_stmt('return')
    assert ast.unparse(stmt1) == 'return None'
    assert ast.unparse(stmt2) == 'tapl_typing.add_return_type(s0, s0.NoneType)'


def test_return2():
    [stmt1, stmt2] = parse_stmt('return True')
    assert ast.unparse(stmt1) == 'return True'
    assert ast.unparse(stmt2) == 'tapl_typing.add_return_type(s0, s0.Bool)'


def test_if():
    [stmt1, stmt2] = parse_stmt('if a == 2:')
    assert ast.unparse(stmt1) == 'if a == 2:'
    assert (
        ast.unparse(stmt2)
        == """
with tapl_typing.scope_forker(s0) as f0:
    s1 = tapl_typing.fork_scope(f0)
    s1.a == s1.Int
    s1 = tapl_typing.fork_scope(f0)
""".strip()
    )


def test_function1():
    [stmt1, stmt2] = parse_module("""
def hello():
    return 0
""")
    assert (
        ast.unparse(stmt1)
        == """
def hello():
    return 0
""".strip()
    )
    assert (
        ast.unparse(stmt2)
        == """
def hello():
    s1 = tapl_typing.create_scope(parent__sa=s0)
    tapl_typing.add_return_type(s1, s1.Int)
    return tapl_typing.get_return_type(s1)
s0.hello = tapl_typing.create_function([], hello())
""".strip()
    )


def test_function_with_return_type():
    [stmt1, stmt2] = parse_module("""
def hello() -> Int:
    return 0
""")
    assert (
        ast.unparse(stmt1)
        == """
def hello():
    return 0
""".strip()
    )
    assert (
        ast.unparse(stmt2)
        == """
def hello():
    s1 = tapl_typing.create_scope(parent__sa=s0)
    s1.hello = tapl_typing.create_function([], s1.Int)
    tapl_typing.set_return_type(s1, s1.Int)
    tapl_typing.add_return_type(s1, s1.Int)
    return tapl_typing.get_return_type(s1)
s0.hello = tapl_typing.create_function([], hello())
""".strip()
    )


def test_area_function_codegen():
    [stmt1, stmt2] = parse_module("""
def area(radius: Int):
    return 3.14 * radius * radius
""")
    assert (
        ast.unparse(stmt1)
        == """
def area(radius):
    return 3.14 * radius * radius
""".strip()
    )
    assert (
        ast.unparse(stmt2)
        == """
def area(radius):
    s1 = tapl_typing.create_scope(parent__sa=s0, radius=radius)
    tapl_typing.add_return_type(s1, s1.Float * s1.radius * s1.radius)
    return tapl_typing.get_return_type(s1)
s0.area = tapl_typing.create_function([s0.Int], area(s0.Int))
""".strip()
    )


def test_untyped_function():
    [stmt1, stmt2] = parse_module("""
def greet(name):
    print('Hello, ' + name)
""")
    assert (
        ast.unparse(stmt1)
        == """
def greet(name):
    print('Hello, ' + name)
""".strip()
    )
    assert (
        ast.unparse(stmt2)
        == """
def greet(name):
    s1 = tapl_typing.create_scope(parent__sa=s0, name=name)
    s1.print(s1.Str + s1.name)
    return tapl_typing.get_return_type(s1)
s0.greet = greet
""".strip()
    )


def test_if_else_stmt():
    [stmt1, stmt2] = parse_module("""
if a == 2:
    b = 7
else:
    b = 'banana'
print(b)
""")
    assert (
        ast.unparse(stmt1)
        == """
if a == 2:
    b = 7
else:
    b = 'banana'
print(b)
""".strip()
    )
    assert (
        ast.unparse(stmt2)
        == """
with tapl_typing.scope_forker(s0) as f0:
    s1 = tapl_typing.fork_scope(f0)
    s1.a == s1.Int
    s1.b = s1.Int
    s1 = tapl_typing.fork_scope(f0)
    s1.b = s1.Str
s0.print(s0.b)
""".strip()
    )


def test_with_stmt():
    [stmt1, stmt2] = parse_module("""
with context() as b:
    b.do_something()
""")
    assert (
        ast.unparse(stmt1)
        == """
with context() as b:
    b.do_something()
""".strip()
    )
    assert (
        ast.unparse(stmt2)
        == """
with s0.context() as s0.b:
    s0.b.do_something()
""".strip()
    )


def test_class1():
    [stmt1, stmt2] = parse_module("""
class Circle:
    def __init__(self, radius: Float):
        self.radius = radius
""")
    assert (
        ast.unparse(stmt1)
        == """
class Circle:

    def __init__(self, radius):
        self.radius = radius
""".strip()
    )
    assert (
        ast.unparse(stmt2)
        == """
class Circle:

    def __init__(self, radius):
        s1 = tapl_typing.create_scope(parent__sa=s0, self=self, radius=radius)
        s1.self.radius = s1.radius
        return tapl_typing.get_return_type(s1)
s0.Circle = tapl_typing.create_class(cls=Circle, init_args=[s0.Float], methods=[])
""".strip()
    )


def test_class2():
    [stmt1, stmt2] = parse_module("""
class Dog:
    def bark(self) -> Str:
        return 'Woof! Woof!'
""")
    assert (
        ast.unparse(stmt1)
        == """
class Dog:

    def bark(self):
        return 'Woof! Woof!'
""".strip()
    )
    assert (
        ast.unparse(stmt2)
        == """
class Dog:

    def bark(self):
        s1 = tapl_typing.create_scope(parent__sa=s0, self=self)
        s1.bark = tapl_typing.create_function([], s1.Str)
        tapl_typing.set_return_type(s1, s1.Str)
        tapl_typing.add_return_type(s1, s1.Str)
        return tapl_typing.get_return_type(s1)
s0.Dog = tapl_typing.create_class(cls=Dog, init_args=[], methods=[('bark', [])])
""".strip()
    )


def test_try_stmt():
    [stmt1, stmt2] = parse_module("""
try:
    do_something()
except SomeError:
    handle_error()
finally:
    cleanup()
""")
    assert (
        ast.unparse(stmt1)
        == """
try:
    do_something()
except SomeError:
    handle_error()
finally:
    cleanup()
""".strip()
    )
    assert (
        ast.unparse(stmt2)
        == """
with tapl_typing.scope_forker(s0) as f0:
    s1 = tapl_typing.fork_scope(f0)
    s1.do_something()
    s1 = tapl_typing.fork_scope(f0)
    s1.handle_error()
    s1 = tapl_typing.fork_scope(f0)
    s1.cleanup()
""".strip()
    )



from tapl_lang.lib import tapl_typing
from tapl_lang.pythonlike.predef1 import predef_scope as predef_scope__sa
s0 = tapl_typing.create_scope(parent__sa=predef_scope__sa)
tapl_typing.import_module(s0, ['tapl_dev'], 'tapl_lang.lib', 0)

def union_print(a):
    s1 = tapl_typing.create_scope(parent__sa=s0, a=a)
    s1.tapl_dev.log(s1.a)
    return tapl_typing.get_return_type(s1)
s0.union_print = tapl_typing.create_function([s0.Int | s0.Str], union_print(s0.Int | s0.Str))
s0.union_print(s0.Int)
s0.union_print(s0.Str)



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import pytest

from tapl_lang.lib import dynamic_attribute


class MySubject(dynamic_attribute.DynamicAttributeMixin):
    def __init__(self):
        self.vars__sa = {}

    def load__sa(self, key):
        try:
            return self.vars__sa[key]
        except KeyError:
            raise AttributeError(f'{self.__class__.__name__} has no attribute "{key}"') from None

    def store__sa(self, key, value):
        self.vars__sa[key] = value

    def delete__sa(self, key):
        del self.vars__sa[key]

    def __repr__(self):
        return f'MySubject{self.vars__sa}'


def test_define_variable():
    a = MySubject()
    a.x = 42
    assert a.x == 42


def test_undefined_variable():
    s = MySubject()
    s.store__sa('a', 100)
    del s.a
    with pytest.raises(AttributeError):
        _ = s.a


def test_repr():
    p = MySubject()
    assert repr(p) == 'MySubject{}'


def test_binop():
    s = MySubject()
    s.store__sa('__add__', lambda other: f'Added {other}')
    assert s + 3 == 'Added 3'


def test_binop_error():
    s = MySubject()
    with pytest.raises(TypeError):
        _ = s + 3



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception


from tapl_lang.core import parser, syntax
from tapl_lang.core.language import Language
from tapl_lang.lib import terms
from tapl_lang.pythonlike import grammar as pythonlike_grammar

IMPORT_LEVEL = 0

GRAMMAR = pythonlike_grammar.get_grammar()


class PythonlikeLanguage(Language):
    def get_grammar(self, parent_stack: list[syntax.Term]) -> parser.Grammar:
        del parent_stack
        return GRAMMAR

    def create_header_for_evaluate_layer(self) -> syntax.Term:
        return syntax.TermList([])

    def create_header_for_typecheck_layer(self) -> syntax.Term:
        location = syntax.Location(start=syntax.Position(line=1, column=0))
        return syntax.TermList(
            [
                terms.ImportFrom(
                    'tapl_lang.lib',
                    [terms.Alias(name='tapl_typing')],
                    IMPORT_LEVEL,
                    location=location,
                ),
                terms.ImportFrom(
                    'tapl_lang.pythonlike.predef1',
                    [terms.Alias(name='predef_scope', asname='predef_scope__sa')],
                    IMPORT_LEVEL,
                    location=location,
                ),
                terms.Assign(
                    targets=[terms.Name(location=location, id=lambda setting: setting.scope_name, ctx='store')],
                    value=terms.Call(
                        location=location,
                        func=terms.Path(
                            location=location,
                            names=['tapl_typing', 'create_scope'],
                            ctx='load',
                            mode=terms.MODE_EVALUATE,
                        ),
                        args=[],
                        keywords=[
                            (
                                'parent__sa',
                                terms.Name(location=location, id='predef_scope__sa', ctx='load'),
                            )
                        ],
                    ),
                    location=location,
                ),
            ]
        )

    def get_predef_headers(self) -> list[syntax.Term]:
        headers: list[syntax.Term] = [
            self.create_header_for_evaluate_layer(),
            self.create_header_for_typecheck_layer(),
        ]
        return headers



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

"""
Type system implementation for TAPL language.

Core Kinds:
- BaseKind: Abstract base class for all types, and types does not have to inherit from it.
- Union: Represents a union of types (T1 | T2)
- Intersection: Represents an intersection of types (T1 & T2)
- Any: Top type - supertype of all types, except NoneType
- Nothing: Bottom type - subtype of all types
- NoneType: Singleton type for None value
- Record: Structural record type with labeled fields
- Function: Function type with positional and named arguments

Type Checking:
- check_subtype(subtype, supertype): Checks if 'subtype' is a subtype of 'supertype'
- check_type_equality(a, b): Checks if two types are equivalent
- Uses caching and assumption stack to handle recursive type definitions

Design Notes:
- All types are considered as immutable
- Naming convention: Variables with __sa suffix indicate internal type system attributes
- Methods is_subtype_of__sa and is_supertype_of__sa return None when inconclusive
- Inspired by Kotlin's type hierarchy
    - https://stackoverflow.com/a/54762815/22663977
    - NoneType is not subtype of Any, but it is a subtype of Union that includes NoneType
    - Any | NoneType is a top type, supertype of all types
    - Nothing is a bottom type, subtype of all types
"""

from tapl_lang.lib import dynamic_attribute


# FIXME: extend from threading.local, and write a unit test for this - https://docs.python.org/3/library/threading.html#thread-local-data
class TypeCheckerState:
    """Holds transient state for subtype checks (cache + assumption stack)."""

    def __init__(self):
        self.cached_subtype_pairs = {}  # mapping of (subtype, supertype) pairs to bool
        self.assumed_subtype_pairs = []  # stack of (subtype, supertype) pairs


_TYPE_CHECKER_STATE = TypeCheckerState()


def compute_subtype(subtype, supertype):
    if hasattr(supertype, 'is_supertype_of__sa') and hasattr(subtype, 'is_subtype_of__sa'):
        is_supertype = supertype.is_supertype_of__sa(subtype)
        is_subtype = subtype.is_subtype_of__sa(supertype)
        # Return Truee if both methods agree on True, or if one is True and the other is inconclusive.
        return (
            (is_supertype and is_subtype)
            or (is_supertype is None and is_subtype)
            or (is_supertype and is_subtype is None)
        )
    # Fallback: treat them as literals and check for equality.
    return subtype == supertype


def check_subtype(subtype, supertype):
    if subtype is supertype:
        return True
    pair = (subtype, supertype)
    result = _TYPE_CHECKER_STATE.cached_subtype_pairs.get(pair)
    if result is not None:
        return result
    if pair in _TYPE_CHECKER_STATE.assumed_subtype_pairs:
        return True
    try:
        _TYPE_CHECKER_STATE.assumed_subtype_pairs.append(pair)
        result = compute_subtype(subtype, supertype)
        _TYPE_CHECKER_STATE.cached_subtype_pairs[pair] = result
    finally:
        _TYPE_CHECKER_STATE.assumed_subtype_pairs.pop()
    return result


def check_type_equality(a, b):
    return check_subtype(a, b) and check_subtype(b, a)


# rename to keep_root_types or take_root_types or similar #mvp
def drop_same_types(types):
    # FIXME: Build a directed graph, and keep only roots of the forests
    result = []
    for t in types:
        for r in result:
            if check_type_equality(t, r):
                break
        else:
            result.append(t)
    return result


class BaseKind(dynamic_attribute.DynamicAttributeMixin):
    def is_supertype_of__sa(self, subtype):
        del subtype  # unused

    def is_subtype_of__sa(self, supertype):
        del supertype  # unused

    def __or__(self, other):
        return create_union(self, other)

    def __repr__(self):
        return 'BaseKind!'


# FIXME: what happens when '|' and '&' operators are used for binary operation instead of type construction?


# Example: alpha <: (alpha | beta) or beta <: (alpha | beta)
class Union(BaseKind):
    def __init__(self, types, title=None):
        if len(types) <= 1:
            raise ValueError('Union requires at least two types.')
        self._types__sa = types
        self._title__sa = title

    def is_supertype_of__sa(self, subtype):
        # Example: alpha <: (alpha | beta)
        if any(check_subtype(subtype, typ) for typ in self._types__sa):
            return True
        # Inconclusive when subtype_ is itself a Union.
        # Example: (alpha | beta) <: (alpha | beta | gamma)
        return None

    def is_subtype_of__sa(self, supertype):
        # Example: (alpha | beta) <: (alpha | beta | gamma)
        return all(check_subtype(typ, supertype) for typ in self._types__sa)

    def __repr__(self):
        if self._title__sa is not None:
            return self._title__sa
        return ' | '.join([str(t) for t in self._types__sa])

    def iter_types__sa(self):
        yield from self._types__sa


# FIXME: implement '&' operator for Intersection
# Example: alpha & beta <: alpha or alpha & beta <: beta
class Intersection(BaseKind):
    def __init__(self, types, title=None):
        if len(types) <= 1:
            raise ValueError('At least two types are required to create Intersection.')
        self._types__sa = types
        self._title__sa = title

    def is_supertype_of__sa(self, subtype):
        # Example: (alpha & beta & gamma) <: (alpha & beta)
        return all(check_subtype(subtype, typ) for typ in self._types__sa)

    def is_subtype_of__sa(self, supertype):
        # Example: (alpha & beta) <: alpha
        if any(check_subtype(typ, supertype) for typ in self._types__sa):
            return True
        # Inconclusive when supertype_ is itself an Intersection.
        # Example: (alpha & beta & gamma) <: (alpha & beta)
        return None

    def __repr__(self):
        if self._title__sa is not None:
            return self._title__sa
        return ' & '.join([str(t) for t in self._types__sa])

    def iter_types__sa(self):
        yield from self._types__sa


# Top type
class Any(BaseKind):
    def is_supertype_of__sa(self, subtype):
        if isinstance(subtype, Any):
            return True
        # Inconclusive, let subtype decide
        return None

    def is_subtype_of__sa(self, supertype):
        if isinstance(supertype, Any):
            return True
        # Inconclusive, example: Any <: (Any | NoneType)
        return None

    def __repr__(self):
        return 'Any'


# Bottom type
class Nothing(BaseKind):
    def is_supertype_of__sa(self, subtype):
        if isinstance(subtype, Nothing):
            return True
        # Inconclusive, (Nothing & NoneType) <: Nothing
        return None

    def is_subtype_of__sa(self, supertype):
        if isinstance(supertype, Nothing):
            return True
        if isinstance(supertype, Any):
            return True
        # Inconclusive, let supertype decide
        return None

    def __repr__(self):
        return 'Nothing'


# Inspired by Kotlin type system - https://stackoverflow.com/a/54762815/22663977
class NoneType(BaseKind):
    def is_supertype_of__sa(self, subtype):
        if isinstance(subtype, NoneType):
            return True
        # Inconclusive, example: (NoneType & T) <: NoneType
        return None

    def is_subtype_of__sa(self, supertype):
        if isinstance(supertype, NoneType):
            return True
        # Inconclusive, example: NoneType <: (T | NoneType)
        return None

    def __repr__(self):
        return 'None'


class Record(BaseKind):
    def __init__(self, fields, label=None):
        self._fields__sa = fields
        self._label__sa = label

    def is_supertype_of__sa(self, subtype):
        if isinstance(subtype, Nothing):
            return True
        # Inconclusive, example: {a: Alpha, b: Beta} <: {a: Alpha}
        return None

    def is_subtype_of__sa(self, supertype):
        if isinstance(supertype, Any):
            return True
        if isinstance(supertype, Record):
            for label in supertype.keys__sa():
                if label not in self._fields__sa:
                    return False
                if not check_subtype(self._fields__sa[label], supertype.try_load(label)):
                    return False
            return True
        # Inconclusive, example: {a: Alpha, b: Beta} <: (Any | NoneType)
        return None

    def keys__sa(self):
        yield from self._fields__sa.keys()

    def fields_repr__sa(self):
        field_strs = [f'{key}: {typ}' for key, typ in self._fields__sa.items()]
        return '{' + ', '.join(field_strs) + '}'

    def __repr__(self):
        if self._label__sa is not None:
            return self._label__sa
        return self.fields_repr__sa()

    def try_load(self, key):
        return self._fields__sa.get(key)

    def load__sa(self, key):
        if key in self._fields__sa:
            return self._fields__sa[key]
        raise AttributeError(f'{self.get_label__sa()} record has no attribute "{key}"')

    def store__sa(self, name: str, value: Any) -> None:
        stored_value = self.load__sa(name)
        if not check_subtype(value, stored_value):
            raise TypeError(f'Type error in variable "{name}": Expected type "{stored_value}", but found "{value}".')

    def get_label__sa(self):
        if self._label__sa is not None:
            return self._label__sa
        return 'Unlabeled record class'


_PAIR_ELEMENT_COUNT = 2


# FIXME: Implement vararg, kwonlyargs, kw_defaults, kwarg, and defaults. For example string.format function.
class Function(BaseKind):
    def __init__(self, posonlyargs, args, result=None, lazy_result=None):
        if not isinstance(posonlyargs, list):
            raise TypeError('Function posonlyargs must be a list.')
        if not isinstance(args, list):
            raise TypeError('Function args must be a list.')
        if any(not (isinstance(arg, tuple) and len(arg) == _PAIR_ELEMENT_COUNT) for arg in args):
            raise ValueError(f'Function args must be a list of (name, type) pairs. args={args}')
        if any(not (isinstance(name, str)) for name, _ in args):
            raise ValueError('Function args must be a list of (str, Any) tuples.')
        if lazy_result is not None and result is not None:
            raise ValueError('Pass either the result or lazy_result argument, but not both.')

        self.posonlyargs__sa = posonlyargs  # list of Type Proxy
        self.args__sa = args  # list of (name, Type Proxy)
        self._result__sa = result
        self._lazy_result__sa = lazy_result

    # FIXME: implement supertype and subtype checking for function types
    def is_supertype_of__sa(self, subtype):
        if isinstance(subtype, Nothing):
            return True
        # Inconclusive, example: ???
        return None

    def is_subtype_of__sa(self, supertype):
        if isinstance(supertype, Any):
            return True
        if isinstance(supertype, Function):
            if len(self.posonlyargs__sa) != len(supertype.posonlyargs__sa):
                return False
            for p_self, p_super in zip(self.posonlyargs__sa, supertype.posonlyargs__sa, strict=False):
                if not check_subtype(p_super, p_self):
                    return False
            if len(self.args__sa) != len(supertype.args__sa):
                return False
            for (n_self, a_self), (n_super, a_super) in zip(self.args__sa, supertype.args__sa, strict=True):
                if n_self != n_super:
                    return False
                if not check_subtype(a_super, a_self):
                    return False
            return check_subtype(self.result__sa, supertype.result__sa)
        # Inconclusive, example: ???
        return None

    def __repr__(self):
        args = [str(t) for t in self.posonlyargs__sa]
        args += [f'{name}: {typ}' for name, typ in self.args__sa]
        args_str = f'({", ".join(args)})'
        if self._lazy_result__sa:
            return f'{args_str}->[uncomputed]'
        return f'{args_str}->{self._result__sa}'

    def apply(self, *arguments):
        actual_all_args = list(arguments)
        expected_args_count = len(self.posonlyargs__sa) + len(self.args__sa)
        if len(actual_all_args) != expected_args_count:
            raise TypeError(f'Expected {expected_args_count} arguments, got {len(actual_all_args)}')
        actual_posonlyargs = actual_all_args[: len(self.posonlyargs__sa)]
        actual_args = actual_all_args[len(self.posonlyargs__sa) :]
        for p, a in zip(self.posonlyargs__sa, actual_posonlyargs, strict=False):
            if not check_subtype(a, p):
                raise TypeError(
                    f'Function positional arguments are not equal: expected={self.posonlyargs__sa} actual={actual_posonlyargs}'
                )
        for p, a in zip(self.args__sa, actual_args, strict=True):
            if not check_subtype(a, p):
                raise TypeError(f'Function arguments are not equal: expected={self.args__sa} actual={actual_args}')
        return self.result__sa

    def load__sa(self, key):
        if key == '__call__':
            return self.apply
        raise AttributeError(f'{self} function has no attribute "{key}"')

    @property
    def result__sa(self):
        self.evaluate_lazy_result__sa()
        return self._result__sa

    def evaluate_lazy_result__sa(self):
        if self._lazy_result__sa:
            self._result__sa = self._lazy_result__sa()
            self._lazy_result__sa = None


def create_union(*args):
    result = []
    for arg in args:
        # Union of unions are flattened
        subject = arg
        if isinstance(subject, Union):
            result.extend(subject.iter_types__sa())  # consume as iterable
        else:
            result.append(arg)
    result = drop_same_types(result)
    # Unions of a single argument vanish
    if len(result) == 1:
        return next(iter(result))
    return Union(types=result)


def create_function(args, result=None, lazy_result=None):
    posonly = []
    regular = []
    tuple_found = False
    for arg in args:
        if isinstance(arg, tuple):
            tuple_found = True
            args.append((arg[0], arg[1]))
        elif not tuple_found:
            posonly.append(arg)
        else:
            raise ValueError('Positional-only arguments must come before regular arguments')
    return Function(posonlyargs=posonly, args=regular, result=result, lazy_result=lazy_result)



from tapl_lang.lib import tapl_typing
from tapl_lang.pythonlike.predef1 import predef_scope as predef_scope__sa
s0 = tapl_typing.create_scope(parent__sa=predef_scope__sa)
tapl_typing.import_module(s0, ['tapl_dev'], 'tapl_lang.lib', 0)

class SimplestClass:
    pass
s0.SimplestClass = tapl_typing.create_class(cls=SimplestClass, init_args=[], methods=[])

def accept(param):
    s1 = tapl_typing.create_scope(parent__sa=s0, param=param)
    pass
    return tapl_typing.get_return_type(s1)
s0.accept = tapl_typing.create_function([s0.SimplestClass.result__sa], accept(s0.SimplestClass.result__sa))
s0.accept(s0.SimplestClass())

class Circle:

    def __init__(self, radius):
        s1 = tapl_typing.create_scope(parent__sa=s0, self=self, radius=radius)
        s1.self.radius = s1.radius
        return tapl_typing.get_return_type(s1)

    def area(self):
        s1 = tapl_typing.create_scope(parent__sa=s0, self=self)
        s1.area = tapl_typing.create_function([], s1.Float)
        tapl_typing.set_return_type(s1, s1.Float)
        tapl_typing.add_return_type(s1, s1.Float * s1.self.radius * s1.self.radius)
        return tapl_typing.get_return_type(s1)

    def set_radius(self, radius):
        s1 = tapl_typing.create_scope(parent__sa=s0, self=self, radius=radius)
        s1.self.radius = s1.radius
        return tapl_typing.get_return_type(s1)

    def is_bigger_than(self, radius):
        s1 = tapl_typing.create_scope(parent__sa=s0, self=self, radius=radius)
        tapl_typing.add_return_type(s1, s1.self.radius > s1.radius)
        return tapl_typing.get_return_type(s1)
s0.Circle = tapl_typing.create_class(cls=Circle, init_args=[s0.Float], methods=[('area', []), ('set_radius', [s0.Float]), ('is_bigger_than', [s0.Float])])

def print_area(circle):
    s1 = tapl_typing.create_scope(parent__sa=s0, circle=circle)
    s1.tapl_dev.log(s1.circle.area())
    s1.tapl_dev.log(s1.circle.is_bigger_than(s1.Float))
    return tapl_typing.get_return_type(s1)
s0.print_area = tapl_typing.create_function([s0.Circle.result__sa], print_area(s0.Circle.result__sa))
s0.print_area(s0.Circle(s0.Float))
